from dataclasses import field
from importlib import import_module
from typing import Dict, List, Optional

from marshmallow import fields, ValidationError

from ludwig.api_annotations import DeveloperAPI
from ludwig.schema import utils as schema_utils
from ludwig.schema.hyperopt import utils as hyperopt_utils
from ludwig.schema.utils import ludwig_dataclass


def points_to_evaluate_field(description: Optional[str] = None) -> fields.Field:
    return schema_utils.DictList(
        description=description
        or (
            "Initial parameter suggestions to be run first. This is for when you already have some good parameters "
            "you want to run first to help the algorithm make better suggestions for future parameters. Needs to be "
            "a list of dicts containing the configurations."
        ),
    )


def evaluated_rewards_field(description: Optional[str] = None) -> fields.Field:
    return schema_utils.List(
        description=description
        or (
            "If you have previously evaluated the parameters passed in as points_to_evaluate you can avoid re-running "
            "those trials by passing in the reward attributes as a list so the optimiser can be told the results "
            "without needing to re-compute the trial. Must be the same length as `points_to_evaluate`."
        )
    )


@DeveloperAPI
@ludwig_dataclass
class BaseSearchAlgorithmConfig(schema_utils.BaseMarshmallowConfig):
    """Basic search algorithm settings."""

    type: str = schema_utils.String(default="hyperopt", description="The search algorithm to use.")

    def set_random_state(self, ludwig_random_state: int) -> None:
        """Overwrite the config random state.

        Search algorithms refer to random state by different names, however we want to overwrite unset random states
        with the Ludwig random state. This method uses a registry of random state field names to provide a single
        interface across all search algorithms.
        """
        rs_field = hyperopt_utils.get_search_algorithm_random_state_field(self.type)
        if rs_field is not None and self.__getattribute__(rs_field) is None:
            self.__setattr__(rs_field, ludwig_random_state)

    def dependencies_installed(self) -> bool:
        """Some search algorithms require additional packages to be installed, check that they are available."""
        missing_packages = []
        missing_installs = []
        for package_name, install_name in hyperopt_utils.get_search_algorithm_dependencies(self.type):
            try:
                import_module(package_name)
            except ImportError:
                missing_packages.append(package_name)
                missing_installs.append(install_name)

        if missing_packages:
            missing_packages = ", ".join(missing_packages)
            missing_installs = " ".join(missing_installs)
            raise ImportError(
                f"Some packages needed to use hyperopt search algorithm {self.type} are not installed: "
                f"{missing_packages}. To add these dependencies, run `pip install {missing_installs}`. For more "
                "details, please refer to Ray Tune documentation for this search algorithm."
            )
        return True


@DeveloperAPI
def SearchAlgorithmDataclassField(description: str = "", default: Dict = {"type": "variant_generator"}):
    class SearchAlgorithmMarshmallowField(fields.Field):
        def _deserialize(self, value, attr, data, **kwargs):
            if isinstance(value, dict):
                try:
                    return BaseSearchAlgorithmConfig.Schema().load(value)
                except (TypeError, ValidationError):
                    raise ValidationError(f"Invalid params for scheduler: {value}, see SearchAlgorithmConfig class.")
            raise ValidationError("Field should be dict")

        def _jsonschema_type_mapping(self):
            return {
                # **schema_utils.unload_jsonschema_from_marshmallow_class(BaseSearchAlgorithmConfig),
                "type": "object",
                "properties": {
                    "type": {
                        "type": "string",
                        "enum": list(hyperopt_utils.search_algorithm_config_registry.keys()),
                        "default": default["type"],
                        "description": "The type of scheduler to use during hyperopt",
                    },
                },
                "title": "search_algorithm_options",
                "required": ["type"],
                "description": description,
            }

    if not isinstance(default, dict):
        raise ValidationError(f"Invalid default: `{default}`")

    load_default = lambda: BaseSearchAlgorithmConfig.Schema().load(default)
    dump_default = BaseSearchAlgorithmConfig.Schema().dump(default)

    return field(
        metadata={
            "marshmallow_field": SearchAlgorithmMarshmallowField(
                allow_none=False,
                load_default=load_default,
                dump_default=dump_default,
                metadata={"description": description, "parameter_metadata": None},
            )
        },
        default_factory=load_default,
    )


@DeveloperAPI
@hyperopt_utils.register_search_algorithm_config("random", random_state_field="random_state")
@hyperopt_utils.register_search_algorithm_config("variant_generator", random_state_field="random_state")
@ludwig_dataclass
class BasicVariantSAConfig(BaseSearchAlgorithmConfig):
    type: str = schema_utils.StringOptions(options=["random", "variant_generator"], default="random", allow_none=False)

    points_to_evaluate: Optional[List[Dict]] = schema_utils.DictList(
        description=(
            "Initial parameter suggestions to be run first. This is for when you already have some good parameters "
            "you want to run first to help the algorithm make better suggestions for future parameters. Needs to be "
            "a list of dicts containing the configurations."
        )
    )

    max_concurrent: int = schema_utils.NonNegativeInteger(
        default=0, description="Maximum number of concurrently running trials. If 0 (default), no maximum is enforced."
    )

    constant_grid_search: bool = schema_utils.Boolean(
        default=False,
        description=(
            "If this is set to True, Ray Tune will first try to sample random values and keep them constant over grid "
            "search parameters. If this is set to False (default), Ray Tune will sample new random parameters in each "
            "grid search condition."
        ),
    )

    random_state: int = schema_utils.Integer(
        default=None,
        allow_none=True,
        description=(
            "Seed or numpy random generator to use for reproducible results. If None (default), will use the global "
            "numpy random generator (np.random). Please note that full reproducibility cannot be guaranteed in a "
            "distributed environment."
        ),
    )


@DeveloperAPI
@hyperopt_utils.register_search_algorithm_config(
    "ax", dependencies=[("ax", "ax-platform"), ("sqlalchemy", "sqlalchemy")]
)
@ludwig_dataclass
class AxSAConfig(BaseSearchAlgorithmConfig):
    type: str = schema_utils.ProtectedString("ax")

    space: Optional[List[Dict]] = schema_utils.DictList(
        description=(
            r"Parameters in the experiment search space. Required elements in the dictionaries are: \“name\” (name of "
            r"this parameter, string), \“type\” (type of the parameter: \“range\”, \“fixed\”, or \“choice\”, string), "
            r"\“bounds\” for range parameters (list of two values, lower bound first), \“values\” for choice "
            r"parameters (list of values), and \“value\” for fixed parameters (single value)."
        )
    )

    points_to_evaluate: Optional[List[Dict]] = points_to_evaluate_field()

    parameter_constraints: Optional[List] = schema_utils.List(
        description=r"Parameter constraints, such as \“x3 >= x4\” or \“x3 + x4 >= 2\”."
    )

    outcome_constraints: Optional[List] = schema_utils.List(
        description=r"Outcome constraints of form \“metric_name >= bound\”, like \“m1 <= 3.\”"
    )


@DeveloperAPI
@hyperopt_utils.register_search_algorithm_config(
    "bayesopt", random_state_field="random_state", dependencies=[("bayes_opt", "bayesian-optimization")]
)
@ludwig_dataclass
class BayesOptSAConfig(BaseSearchAlgorithmConfig):
    type: str = schema_utils.ProtectedString("bayesopt")

    space: Optional[Dict] = schema_utils.Dict(
        description=(
            "Continuous search space. Parameters will be sampled from this space which will be used to run trials"
        )
    )

    points_to_evaluate: Optional[List[Dict]] = points_to_evaluate_field()

    utility_kwargs: Optional[Dict] = schema_utils.Dict(
        description=(
            "Parameters to define the utility function. The default value is a dictionary with three keys: "
            "- kind: ucb (Upper Confidence Bound) - kappa: 2.576 - xi: 0.0"
        )
    )

    random_state: int = schema_utils.Integer(default=None, allow_none=True, description="Used to initialize BayesOpt.")

    random_search_steps: int = schema_utils.Integer(
        default=10,
        description=(
            "Number of initial random searches. This is necessary to avoid initial local overfitting of "
            "the Bayesian process."
        ),
    )

    verbose: int = schema_utils.IntegerOptions(
        options=[0, 1, 2], default=0, description="The level of verbosity. `0` is least verbose, `2` is most verbose."
    )

    patience: int = schema_utils.NonNegativeInteger(
        default=5, description="Number of epochs to wait for a change in the top models."
    )

    skip_duplicate: bool = schema_utils.Boolean(
        default=True,
        description=(
            "If False, the optimizer will allow duplicate points to be registered. This behavior may be desired in "
            "high noise situations where repeatedly probing the same point will give different answers. In other "
            "situations, the acquisition may occasionaly generate a duplicate point."
        ),
    )


@DeveloperAPI
@hyperopt_utils.register_search_algorithm_config("blendsearch", dependencies=[("flaml", "flaml[blendsearch]")])
@ludwig_dataclass
class BlendsearchSAConfig(BaseSearchAlgorithmConfig):
    type: str = schema_utils.ProtectedString("blendsearch")


@DeveloperAPI
@hyperopt_utils.register_search_algorithm_config(
    "bohb", random_state_field="seed", dependencies=[("hpbandster", "hpbandster"), ("ConfigSpace", "ConfigSpace")]
)
@ludwig_dataclass
class BOHBSAConfig(BaseSearchAlgorithmConfig):
    type: str = schema_utils.ProtectedString("bohb")

    space: Optional[Dict] = schema_utils.Dict(
        description=(
            "Continuous ConfigSpace search space. Parameters will be sampled from this space which will be used "
            "to run trials."
        )
    )

    bohb_config: Optional[Dict] = schema_utils.Dict(description="configuration for HpBandSter BOHB algorithm")

    points_to_evaluate: Optional[List[Dict]] = points_to_evaluate_field()

    seed: Optional[int] = schema_utils.Integer(
        default=None,
        allow_none=True,
        description=(
            "Optional random seed to initialize the random number generator. Setting this should lead to identical "
            "initial configurations at each run."
        ),
    )

    max_concurrent: int = schema_utils.Integer(
        default=0,
        description=(
            "Number of maximum concurrent trials. If this Searcher is used in a `ConcurrencyLimiter`, the "
            "`max_concurrent` value passed to it will override the value passed here. Set to <= 0 for no limit on "
            "concurrency."
        ),
    )


@DeveloperAPI
@hyperopt_utils.register_search_algorithm_config("cfo", dependencies=[("flaml", "flaml")])
@ludwig_dataclass
class CFOSAConfig(BaseSearchAlgorithmConfig):
    type: str = schema_utils.ProtectedString("cfo")


@DeveloperAPI
@hyperopt_utils.register_search_algorithm_config(
    "dragonfly", random_state_field="random_state_seed", dependencies=[("dragonfly", "dragonfly-opt")]
)
@ludwig_dataclass
class DragonflySAConfig(BaseSearchAlgorithmConfig):
    type: str = schema_utils.ProtectedString("dragonfly")

    optimizer: Optional[str] = schema_utils.StringOptions(
        options=["random", "bandit", "genetic"],
        default=None,
        allow_none=True,
        description=(
            "Optimizer provided from dragonfly. Choose an optimiser that extends `BlackboxOptimiser`. If this is a "
            "string, `domain` must be set and `optimizer` must be one of [random, bandit, genetic]."
        ),
    )

    domain: Optional[str] = schema_utils.StringOptions(
        options=["cartesian", "euclidean"],
        default=None,
        allow_none=True,
        description=(
            "Optional domain. Should only be set if you don't pass an optimizer as the `optimizer` argument. If set, "
            "has to be one of `[cartesian, euclidean]`."
        ),
    )

    space: Optional[List[Dict]] = schema_utils.DictList(
        description=(
            "Search space. Should only be set if you don't pass an optimizer as the `optimizer` argument. Defines the "
            "search space and requires a `domain` to be set. Can be automatically converted from the `param_space` "
            "dict passed to `tune.Tuner()`."
        )
    )

    points_to_evaluate: Optional[List[Dict]] = points_to_evaluate_field()

    evaluated_rewards: Optional[List] = evaluated_rewards_field()

    random_state_seed: Optional[int] = schema_utils.Integer(
        default=None,
        allow_none=True,
        description=(
            "Seed for reproducible results. Defaults to None. Please note that setting this to a value will change "
            "global random state for `numpy` on initalization and loading from checkpoint."
        ),
    )


@DeveloperAPI
@hyperopt_utils.register_search_algorithm_config(
    "hebo", random_state_field="random_state_seed", dependencies=[("hebo", "HEBO")]
)
@ludwig_dataclass
class HEBOSAConfig(BaseSearchAlgorithmConfig):
    type: str = schema_utils.ProtectedString("hebo")

    space: Optional[List[Dict]] = schema_utils.DictList(
        description="A dict mapping parameter names to Tune search spaces or a HEBO DesignSpace object."
    )

    points_to_evaluate: Optional[List[Dict]] = points_to_evaluate_field()

    evaluated_rewards: Optional[List] = evaluated_rewards_field()

    random_state_seed: Optional[int] = schema_utils.Integer(
        default=None,
        allow_none=True,
        description=(
            "Seed for reproducible results. Defaults to None. Please note that setting this to a value will change "
            "global random state for `numpy` on initalization and loading from checkpoint."
        ),
    )

    max_concurrent: int = schema_utils.NonNegativeInteger(
        default=8,
        description=(
            "Number of maximum concurrent trials. If this Searcher is used in a `ConcurrencyLimiter`, the "
            "`max_concurrent` value passed to it will override the value passed here."
        ),
    )


@DeveloperAPI
@hyperopt_utils.register_search_algorithm_config(
    "hyperopt", random_state_field="random_state_seed", dependencies=[("hyperopt", "hyperopt")]
)
@ludwig_dataclass
class HyperoptSAConfig(BaseSearchAlgorithmConfig):
    type: str = schema_utils.ProtectedString("hyperopt")

    space: Optional[List[Dict]] = schema_utils.DictList(
        description=(
            "HyperOpt configuration. Parameters will be sampled from this configuration and will be used to override "
            "parameters generated in the variant generation process."
        )
    )

    points_to_evaluate: Optional[List[Dict]] = points_to_evaluate_field()

    n_initial_points: int = schema_utils.PositiveInteger(
        default=20,
        description=(
            "The number of random evaluations of the objective function before starting to approximate it with tree "
            "parzen estimators. Defaults to 20."
        ),
    )

    random_state_seed: Optional[int] = schema_utils.Integer(
        default=None,
        allow_none=True,
        description=("Seed for reproducible results. Defaults to None."),
    )

    gamma: float = schema_utils.FloatRange(
        min=0.0,
        max=1.0,
        default=0.25,
        description=(
            "The split to use in TPE. TPE models two splits of the evaluated hyperparameters: the top performing "
            "`gamma` percent, and the remaining examples. For more details, see [Making a Science of Model Search: "
            "Hyperparameter Optimization in Hundreds of Dimensions for Vision Architectures.]"
            "(http://proceedings.mlr.press/v28/bergstra13.pdf)."
        ),
    )


@DeveloperAPI
@hyperopt_utils.register_search_algorithm_config("nevergrad", dependencies=[("nevergrad", "nevergrad")])
@ludwig_dataclass
class NevergradSAConfig(BaseSearchAlgorithmConfig):
    type: str = schema_utils.ProtectedString("nevergrad")

    # TODO: Add a registry mapping string names to nevergrad optimizers
    # optimizer: Optional[str] = None

    # TODO: Add schemas for nevergrad optimizer kwargs
    optimizer_kwargs: Optional[Dict] = schema_utils.Dict(
        description="Kwargs passed in when instantiating the optimizer."
    )

    space: Optional[List[Dict]] = schema_utils.DictList(
        description=(
            "Nevergrad parametrization to be passed to optimizer on instantiation, or list of parameter names if you "
            "passed an optimizer object."
        )
    )

    points_to_evaluate: Optional[List[Dict]] = points_to_evaluate_field()


@DeveloperAPI
@hyperopt_utils.register_search_algorithm_config(
    "optuna", random_state_field="seed", dependencies=[("optuna", "optuna")]
)
@ludwig_dataclass
class OptunaSAConfig(BaseSearchAlgorithmConfig):
    type: str = schema_utils.ProtectedString("optuna")

    space: Optional[Dict] = schema_utils.Dict(
        description=(
            "Hyperparameter search space definition for Optuna's sampler. This can be either a dict with parameter "
            "names as keys and optuna.distributions as values, or a Callable - in which case, it should be a "
            "define-by-run function using optuna.trial to obtain the hyperparameter values. The function should "
            "return either a dict of constant values with names as keys, or None. For more information, see "
            "[the Optuna docs]"
            "(https://optuna.readthedocs.io/en/stable/tutorial/10_key_features/002_configurations.html)."
        )
    )

    points_to_evaluate: Optional[List[Dict]] = points_to_evaluate_field()

    # TODO: Add a registry of Optuna samplers schemas
    # sampler = None

    seed: Optional[int] = schema_utils.Integer(
        default=None,
        allow_none=True,
        description=(
            "Seed to initialize sampler with. This parameter is only used when `sampler=None`. In all other cases, "
            "the sampler you pass should be initialized with the seed already."
        ),
    )

    evaluated_rewards: Optional[List] = evaluated_rewards_field()


@DeveloperAPI
@hyperopt_utils.register_search_algorithm_config("skopt", dependencies=[("skopt", "scikit-optimize")])
class SkoptSAConfig(BaseSearchAlgorithmConfig):
    type: str = schema_utils.ProtectedString("skopt")

    optimizer = None

    space: Optional[Dict] = schema_utils.Dict(
        description=(
            "A dict mapping parameter names to valid parameters, i.e. tuples for numerical parameters and lists "
            "for categorical parameters. If you passed an optimizer instance as the optimizer argument, this should "
            "be a list of parameter names instead."
        )
    )

    points_to_evaluate: Optional[List[Dict]] = points_to_evaluate_field()

    evaluated_rewards: Optional[List] = evaluated_rewards_field(
        description=(
            "If you have previously evaluated the parameters passed in as points_to_evaluate you can avoid "
            "re-running those trials by passing in the reward attributes as a list so the optimiser can be told the "
            "results without needing to re-compute the trial. Must be the same length as points_to_evaluate. (See "
            "tune/examples/skopt_example.py)"
        )
    )

    convert_to_python: bool = schema_utils.Boolean(
        default=True,
        description="SkOpt outputs numpy primitives (e.g. `np.int64`) instead of Python types. If this setting is set "
        "to `True`, the values will be converted to Python primitives.",
    )


@DeveloperAPI
@hyperopt_utils.register_search_algorithm_config("zoopt", dependencies=[("zoopt", "zoopt")])
@ludwig_dataclass
class ZooptSAConfig(BaseSearchAlgorithmConfig):
    type: str = schema_utils.ProtectedString("zoopt")

    algo: str = schema_utils.ProtectedString(
        pstring="asracos",
        description="To specify an algorithm in zoopt you want to use. Only support ASRacos currently.",
    )

    budget: Optional[int] = schema_utils.PositiveInteger(
        default=None, allow_none=True, description="Optional. Number of samples."
    )

    dim_dict: Optional[Dict] = schema_utils.Dict(
        description=(
            "Dimension dictionary. For continuous dimensions: (continuous, search_range, precision); For discrete "
            "dimensions: (discrete, search_range, has_order); For grid dimensions: (grid, grid_list). More details "
            "can be found in zoopt package."
        )
    )

    points_to_evaluate: Optional[List[Dict]] = points_to_evaluate_field()

    parallel_num: int = schema_utils.PositiveInteger(
        default=1,
        description=(
            "How many workers to parallel. Note that initial phase may start less workers than this number. More "
            "details can be found in zoopt package."
        ),
    )



import logging
import os
import time
from typing import Callable, Dict, List, Optional, Union

from torch.utils.tensorboard import SummaryWriter

from ludwig.constants import MINIMUM_BATCH_SIZE, TEST, TRAINING, VALIDATION
from ludwig.data.dataset.base import Dataset
from ludwig.distributed.base import DistributedStrategy, LocalStrategy
from ludwig.features.feature_utils import LudwigFeatureDict
from ludwig.models.llm import LLM
from ludwig.models.predictor import LlmFineTunePredictor, LlmPredictor
from ludwig.modules.metric_modules import get_initial_validation_value
from ludwig.schema.trainer import BaseTrainerConfig, FineTuneTrainerConfig, NoneTrainerConfig
from ludwig.trainers.base import BaseTrainer
from ludwig.trainers.registry import register_llm_ray_trainer, register_llm_trainer
from ludwig.trainers.trainer import Trainer
from ludwig.types import ModelConfigDict
from ludwig.utils import time_utils
from ludwig.utils.defaults import default_random_seed
from ludwig.utils.metric_utils import TrainerMetric
from ludwig.utils.metrics_printed_table import print_metrics_table
from ludwig.utils.misc_utils import set_random_seed
from ludwig.utils.torch_utils import get_torch_device
from ludwig.utils.trainer_utils import append_metrics, get_new_progress_tracker, ProgressTracker

logger = logging.getLogger(__name__)


@register_llm_trainer("none")
@register_llm_ray_trainer("none")
class NoneTrainer(BaseTrainer):
    """NoneTrainer is a trainer that does not train a model, only runs evaluation."""

    def __init__(
        self,
        config: NoneTrainerConfig,
        model: LLM,
        resume: float = False,
        skip_save_model: bool = False,
        skip_save_progress: bool = False,
        skip_save_log: bool = False,
        callbacks: List = None,
        report_tqdm_to_ray=False,
        random_seed: float = default_random_seed,
        distributed: Optional[DistributedStrategy] = None,
        device: Optional[str] = None,
        **kwargs,
    ):
        """
        :param config: `ludwig.schema.trainer.NoneTrainerConfig` instance that specifies training hyperparameters
        (default: `ludwig.schema.trainer.NoneTrainerConfig()`).
        :param model: Underlying Ludwig model
        :type model: `ludwig.models.llm.LLM`
        :param resume: Resume training a model that was being trained. (default: False).
        :type resume: Boolean
        :param skip_save_model: Disables saving model weights and hyperparameters each time the model improves. By
                default Ludwig saves model weights after each round of evaluation the validation metric (improves, but
                if the model is really big that can be time consuming. If you do not want to keep the weights and just
                find out what performance a model can get with a set of hyperparameters, use this parameter to skip it,
                but the model will not be loadable later on. (default: False).
        :type skip_save_model: Boolean
        :param skip_save_progress: Disables saving progress each round of evaluation. By default Ludwig saves weights
                and stats after each round of evaluation for enabling resuming of training, but if the model is really
                big that can be time consuming and will uses twice as much space, use this parameter to skip it, but
                training cannot be resumed later on. (default: False).
        :type skip_save_progress: Boolean
        :param skip_save_log: Disables saving TensorBoard logs. By default Ludwig saves logs for the TensorBoard, but if
                it is not needed turning it off can slightly increase the overall speed. (default: False).
        :type skip_save_log: Boolean
        :param callbacks: List of `ludwig.callbacks.Callback` objects that provide hooks into the Ludwig pipeline.
                (default: None).
        :type callbacks: list
        :param report_tqdm_to_ray: Enables using the ray based tqdm Callback for progress bar reporting
        :param random_seed: Default initialization for the random seeds (default: 42).
        :type random_seed: Float
        :param distributed: Distributed strategy (default: None).
        :type distributed: `DistributedStrategy`
        :param device: Device to load the model on from a saved checkpoint (default: None).
        :type device: str
        """

        super().__init__()
        self.config = config
        self.distributed = distributed if distributed is not None else LocalStrategy()
        self.skip_save_log = skip_save_log
        self.resume = resume
        self.skip_save_model = skip_save_model
        self.skip_save_progress = skip_save_progress
        self.random_seed = random_seed
        self.callbacks = callbacks or []
        self.report_tqdm_to_ray = report_tqdm_to_ray

        self.device = device if device is not None else get_torch_device()
        self.model = model.to_device(self.device)
        self.model.metrics_to_device(self.device)

        # Since we are only running evaluation without training, set the model to evaluation mode.
        self.model.eval()

        self.batch_size = self.config.batch_size
        self.eval_batch_size = self.config.eval_batch_size
        self.base_learning_rate = self.config.base_learning_rate
        self.should_shuffle = self.config.should_shuffle
        self.epochs = self.config.epochs
        self.train_steps = self.config.train_steps
        self.steps_per_checkpoint = self.config.steps_per_checkpoint
        self.checkpoints_per_epoch = self.config.checkpoints_per_epoch
        self.early_stop = self.config.early_stop
        self.evaluate_training_set = self.config.evaluate_training_set
        self.skip_all_evaluation = self.config.skip_all_evaluation

    def close_writers(
        self, progress_tracker, save_path, train_summary_writer, validation_summary_writer, test_summary_writer
    ):
        # ================ Finished Training ================
        self.callback(
            lambda c: c.on_trainer_train_teardown(self, progress_tracker, save_path, self.is_coordinator()),
            coordinator_only=False,
        )

        if train_summary_writer is not None:
            train_summary_writer.close()
        if validation_summary_writer is not None:
            validation_summary_writer.close()
        if test_summary_writer is not None:
            test_summary_writer.close()

    def train(
        self,
        training_set: Dataset,
        validation_set: Optional[Dataset] = None,
        test_set: Optional[Dataset] = None,
        save_path: str = "model",
        return_state_dict: bool = False,
        **kwargs,
    ):
        output_features = self.model.output_features

        # ====== Setup file names =======
        tensorboard_log_dir = None
        if self.is_coordinator():
            os.makedirs(save_path, exist_ok=True)
            tensorboard_log_dir = os.path.join(save_path, "logs")

        self.callback(
            lambda c: c.on_trainer_train_setup(self, save_path, self.is_coordinator()), coordinator_only=False
        )

        train_summary_writer = None
        validation_summary_writer = None
        test_summary_writer = None
        if self.is_coordinator() and not self.skip_save_log and tensorboard_log_dir:
            train_summary_writer = SummaryWriter(os.path.join(tensorboard_log_dir, TRAINING))
            if validation_set is not None and validation_set.size > 0:
                validation_summary_writer = SummaryWriter(os.path.join(tensorboard_log_dir, VALIDATION))
            if test_set is not None and test_set.size > 0:
                test_summary_writer = SummaryWriter(os.path.join(tensorboard_log_dir, TEST))

        set_random_seed(self.random_seed)

        progress_tracker = get_new_progress_tracker(
            batch_size=self.batch_size,
            learning_rate=self.base_learning_rate,
            best_eval_metric_value=get_initial_validation_value(self.validation_metric),
            best_increase_batch_size_eval_metric=float("inf"),
            output_features=output_features,
        )

        # When running with Ray, we only need to return the state dict, as it's faster and cheaper to send the
        # state dict over the network than to load the model state here, serialize it back to a state dict, then
        # load it back on the head node.
        return_value = self.model if not return_state_dict else self.model.cpu().state_dict()

        if self.skip_all_evaluation:
            self.close_writers(
                progress_tracker, save_path, train_summary_writer, validation_summary_writer, test_summary_writer
            )
            return (
                return_value,
                progress_tracker.train_metrics,
                progress_tracker.validation_metrics,
                progress_tracker.test_metrics,
            )

        try:
            self.run_evaluation(
                training_set,
                validation_set,
                test_set,
                progress_tracker,
                train_summary_writer,
                validation_summary_writer,
                test_summary_writer,
                output_features,
                save_path,
            )
        finally:
            self.close_writers(
                progress_tracker, save_path, train_summary_writer, validation_summary_writer, test_summary_writer
            )

        return (
            return_value,
            progress_tracker.train_metrics,
            progress_tracker.validation_metrics,
            progress_tracker.test_metrics,
        )

    def train_online(
        self,
        dataset,
    ):
        pass

    def tune_batch_size(
        self,
        config: ModelConfigDict,
        training_set: Dataset,
        random_seed: int = default_random_seed,
        max_trials: int = 20,
        halving_limit: int = 3,
        snapshot_weights: bool = True,
        on_best_batch_size_updated: Optional[Callable[[int, float, int], None]] = None,
        tune_for_training: bool = True,
    ) -> int:
        # TODO: Implement batch size tuning for LLM, currently just returns the default batch size
        # Compared to ECD, this just requires forward passes till we OOM.
        # https://github.com/ludwig-ai/ludwig/issues/3525
        return MINIMUM_BATCH_SIZE

    @property
    def validation_field(self):
        return self.config.validation_field

    @property
    def validation_metric(self):
        return self.config.validation_metric

    # Remote implementations may override this
    def shutdown(self):
        pass

    @property
    def local_rank(self) -> int:
        return 0

    def barrier(self):
        pass

    # Functions needed to treat Trainer as a context manager
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.shutdown()

    @staticmethod
    def get_schema_cls() -> BaseTrainerConfig:
        return NoneTrainerConfig

    def is_coordinator(self) -> bool:
        return self.distributed.rank() == 0

    def callback(self, fn, coordinator_only=True):
        if not coordinator_only or self.is_coordinator():
            for callback in self.callbacks:
                fn(callback)

    def evaluation(
        self,
        dataset: "Dataset",  # noqa: F821
        dataset_name: str,
        metrics_log: Dict[str, Dict[str, List[TrainerMetric]]],
        batch_size: int,
        progress_tracker: ProgressTracker,
    ):
        predictor = LlmPredictor(
            self.model, batch_size=batch_size, distributed=self.distributed, report_tqdm_to_ray=self.report_tqdm_to_ray
        )
        metrics, _ = predictor.batch_evaluation(dataset, collect_predictions=False, dataset_name=dataset_name)

        return append_metrics(self.model, dataset_name, metrics, metrics_log, progress_tracker)

    @classmethod
    def write_eval_summary(
        cls,
        summary_writer,
        metrics,
        step,
    ):
        if not summary_writer:
            return

        for feature_name, output_feature in metrics.items():
            for metric_name, metrics in output_feature.items():
                if metrics:
                    metric_tag = f"{feature_name}/epoch_{metric_name}"
                    metric_val = metrics[-1][-1]
                    summary_writer.add_scalar(metric_tag, metric_val, global_step=step)
        summary_writer.flush()

    def run_evaluation(
        self,
        training_set: Union["Dataset", "RayDataset"],  # noqa: F821
        validation_set: Optional[Union["Dataset", "RayDataset"]],  # noqa: F821
        test_set: Optional[Union["Dataset", "RayDataset"]],  # noqa: F821
        progress_tracker: ProgressTracker,
        train_summary_writer: SummaryWriter,
        validation_summary_writer: SummaryWriter,
        test_summary_writer: SummaryWriter,
        output_features: LudwigFeatureDict,
        save_path: str,
    ) -> bool:
        """Runs evaluation over training, validation, and test sets.

        Also:
        - Prints results, saves results to the progress tracker.
        - Saves the model if the validation score is the best so far
        - If there is no validation set, the model is always saved.

        Returns whether the trainer should early stop, based on validation metrics history.
        """
        start_time = time.time()
        self.callback(lambda c: c.on_eval_start(self, progress_tracker, save_path))

        progress_tracker.checkpoint_number += 1
        if self.is_coordinator():
            logger.info(f"\nRunning evaluation for step: {progress_tracker.steps}, epoch: {progress_tracker.epoch}")

        # ================ Eval ================
        # Run a separate pass over the training data to compute metrics
        # Appends results to progress_tracker.train_metrics.
        if self.evaluate_training_set:
            self.evaluation(
                training_set, "train", progress_tracker.train_metrics, self.eval_batch_size, progress_tracker
            )

        self.write_eval_summary(
            summary_writer=train_summary_writer,
            metrics=progress_tracker.train_metrics,
            step=progress_tracker.steps,
        )

        if validation_set is not None:
            self.callback(lambda c: c.on_validation_start(self, progress_tracker, save_path))

            # eval metrics on validation set
            self.evaluation(
                validation_set,
                VALIDATION,
                progress_tracker.validation_metrics,
                self.eval_batch_size,
                progress_tracker,
            )

            self.write_eval_summary(
                summary_writer=validation_summary_writer,
                metrics=progress_tracker.validation_metrics,
                step=progress_tracker.steps,
            )

            self.callback(lambda c: c.on_validation_end(self, progress_tracker, save_path))

        if test_set is not None:
            self.callback(lambda c: c.on_test_start(self, progress_tracker, save_path))

            # eval metrics on test set
            self.evaluation(test_set, TEST, progress_tracker.test_metrics, self.eval_batch_size, progress_tracker)

            self.write_eval_summary(
                summary_writer=test_summary_writer,
                metrics=progress_tracker.test_metrics,
                step=progress_tracker.steps,
            )

            self.callback(lambda c: c.on_test_end(self, progress_tracker, save_path))

        elapsed_time = (time.time() - start_time) * 1000.0

        if self.is_coordinator():
            logger.info(f"Evaluation took {time_utils.strdelta(elapsed_time)}\n")
            print_metrics_table(
                output_features,
                progress_tracker.train_metrics,
                progress_tracker.validation_metrics,
                progress_tracker.test_metrics,
            )

        # Trigger eval end callback after any model weights save for complete checkpoint
        self.callback(lambda c: c.on_eval_end(self, progress_tracker, save_path))

        return False


@register_llm_trainer("finetune")
class FineTuneTrainer(Trainer):
    @staticmethod
    def get_schema_cls():
        return FineTuneTrainerConfig

    def __init__(
        self,
        config: FineTuneTrainerConfig,
        model: LLM,
        resume: float = False,
        skip_save_model: bool = False,
        skip_save_progress: bool = False,
        skip_save_log: bool = False,
        callbacks: List = None,
        report_tqdm_to_ray=False,
        random_seed: float = default_random_seed,
        distributed: Optional[DistributedStrategy] = None,
        device: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(
            config,
            model,
            resume,
            skip_save_model,
            skip_save_progress,
            skip_save_log,
            callbacks,
            report_tqdm_to_ray,
            random_seed,
            distributed,
            device,
            **kwargs,
        )

    def evaluation(self, dataset, dataset_name, metrics_log, batch_size, progress_tracker):
        predictor = LlmFineTunePredictor(
            self.model, batch_size=batch_size, distributed=self.distributed, report_tqdm_to_ray=self.report_tqdm_to_ray
        )
        metrics, _ = predictor.batch_evaluation(dataset, collect_predictions=False, dataset_name=dataset_name)

        return append_metrics(self.model, dataset_name, metrics, metrics_log, progress_tracker)


class RemoteLLMTrainer(NoneTrainer):
    def __init__(self, gpus=None, gpu_memory_limit=None, allow_parallel_threads=True, **kwargs):
        super().__init__(**kwargs)

        # Only return results from rank 0 to reduce network overhead
        self.train = self.distributed.return_first(self.train)
        self.train_online = self.distributed.return_first(self.train_online)


class RemoteLLMFineTuneTrainer(FineTuneTrainer):
    def __init__(self, gpus=None, gpu_memory_limit=None, allow_parallel_threads=True, **kwargs):
        super().__init__(**kwargs)

        # Only return results from rank 0 to reduce network overhead
        self.train = self.distributed.return_first(self.train)
        self.train_online = self.distributed.return_first(self.train_online)



import math

import numpy as np
from torch.optim import SGD

from ludwig.features.number_feature import NumberInputFeature, NumberOutputFeature
from ludwig.modules.lr_scheduler import LRScheduler
from ludwig.schema.encoders.base import DenseEncoderConfig
from ludwig.schema.features.number_feature import ECDNumberOutputFeatureConfig, NumberInputFeatureConfig
from ludwig.schema.lr_scheduler import LRSchedulerConfig
from ludwig.utils.metric_utils import TrainerMetric
from ludwig.utils.trainer_utils import get_new_progress_tracker


def test_lr_scheduler_warmup_decay():
    total_steps = 10000
    steps_per_checkpoint = 1000
    base_lr = 1.0
    warmup_fraction = 0.1

    module = NumberInputFeature(NumberInputFeatureConfig(name="num1", encoder=DenseEncoderConfig()))

    const_optimizer = SGD(module.parameters(), lr=base_lr)
    const_config = LRSchedulerConfig(warmup_evaluations=0)
    const_scheduler = LRScheduler(
        config=const_config,
        optimizer=const_optimizer,
        steps_per_checkpoint=steps_per_checkpoint,
        total_steps=total_steps,
    )

    linear_optimizer = SGD(module.parameters(), lr=base_lr)
    linear_config = LRSchedulerConfig(warmup_fraction=warmup_fraction, decay="linear")
    linear_scheduler = LRScheduler(
        config=linear_config,
        optimizer=linear_optimizer,
        steps_per_checkpoint=steps_per_checkpoint,
        total_steps=total_steps,
    )

    exp_optimizer = SGD(module.parameters(), lr=base_lr)
    exp_config = LRSchedulerConfig(warmup_fraction=warmup_fraction, decay="exponential")
    exp_scheduler = LRScheduler(
        config=exp_config, optimizer=exp_optimizer, steps_per_checkpoint=steps_per_checkpoint, total_steps=total_steps
    )

    cosine_optimizer = SGD(module.parameters(), lr=base_lr)
    cosine_config = LRSchedulerConfig(warmup_fraction=warmup_fraction, decay="cosine", t_0=steps_per_checkpoint)
    cosine_scheduler = LRScheduler(
        config=cosine_config,
        optimizer=cosine_optimizer,
        steps_per_checkpoint=steps_per_checkpoint,
        total_steps=total_steps,
    )

    warmup_steps = total_steps * warmup_fraction
    for i in range(total_steps):
        # Offset by 1
        step = i + 1

        const_scheduler.step()
        const_lr = const_optimizer.param_groups[0]["lr"]
        assert const_lr == base_lr, f"step: {step}"

        linear_scheduler.step()
        linear_lr = linear_optimizer.param_groups[0]["lr"]

        exp_scheduler.step()
        exp_lr = exp_optimizer.param_groups[0]["lr"]

        cosine_scheduler.step()
        cosine_lr = cosine_optimizer.param_groups[0]["lr"]

        if step < warmup_steps:
            assert linear_lr == exp_lr, f"step: {step}"
            assert linear_lr == cosine_lr, f"step: {step}"
            assert linear_lr < base_lr, f"step: {step}"
        elif step == warmup_steps:
            assert linear_lr == base_lr, f"step: {step}"
            assert cosine_lr == base_lr, f"step: {step}"
            assert exp_lr < base_lr, f"step: {step}"
        else:
            assert linear_lr < base_lr, f"step: {step}"
            assert exp_lr < base_lr, f"step: {step}"
            assert cosine_lr <= base_lr, f"step: {step}"

    assert linear_lr < exp_lr
    assert exp_lr < cosine_lr
    assert cosine_lr == base_lr


def test_lr_scheduler_reduce_on_plateau():
    total_eval_steps = 100
    base_lr = 1.0
    reduce_limit = 3

    module = NumberInputFeature(NumberInputFeatureConfig(name="num1", encoder=DenseEncoderConfig()))
    output1 = NumberOutputFeature(ECDNumberOutputFeatureConfig(name="output1", input_size=10), output_features={})

    optimizer = SGD(module.parameters(), lr=base_lr)
    config = LRSchedulerConfig(
        warmup_evaluations=0,
        decay=None,
        reduce_on_plateau=reduce_limit,
        reduce_on_plateau_patience=10,
        reduce_on_plateau_rate=0.1,
    )
    scheduler = LRScheduler(config=config, optimizer=optimizer, steps_per_checkpoint=0, total_steps=0)

    progress_tracker = get_new_progress_tracker(
        batch_size=64,
        best_eval_metric_value=float("inf"),
        best_increase_batch_size_eval_metric=float("inf"),
        learning_rate=base_lr,
        output_features={"output1": output1},
    )

    num_reductions = 0

    last_lr = optimizer.param_groups[0]["lr"]
    steps_to_plateau = 5
    loss = 10.0
    for epoch in range(total_eval_steps):
        for i in range(100):
            # Simulate batch-wise steps. If we make a mistake, then this will reset
            # the learning rate.
            scheduler.step()

        steps_to_plateau -= 1
        if steps_to_plateau > 0:
            loss -= 0.1

        progress_tracker.train_metrics["output1"]["loss"].append(
            TrainerMetric(epoch=epoch, step=epoch * 100, value=loss)
        )
        scheduler.eval_step(progress_tracker, "output1")
        lr = optimizer.param_groups[0]["lr"]
        if lr < last_lr:
            # Reset steps to plateau
            steps_to_plateau = 5
            num_reductions += 1
        last_lr = lr

    assert num_reductions == reduce_limit

    # 3 reductions that multiply by 0.1 each time
    assert np.isclose(lr, 0.001)


def test_lr_scheduler_cosine_decay_fixed_period():
    total_steps = 10000
    steps_per_checkpoint = 1000
    base_lr = 1.0

    module = NumberInputFeature(NumberInputFeatureConfig(name="num1", encoder=DenseEncoderConfig()))

    optimizer = SGD(module.parameters(), lr=base_lr)
    config = LRSchedulerConfig(decay="cosine", t_0=steps_per_checkpoint, decay_rate=0, reduce_on_plateau=0)
    scheduler = LRScheduler(config=config, optimizer=optimizer, steps_per_checkpoint=0, total_steps=0)

    curr_lr = base_lr
    prev_lr = base_lr
    num_restarts = 0
    for step in range(total_steps + 1):
        # Cosine annealing formula
        expected_lr = base_lr * 0.5 * (1 + math.cos(math.pi * (step % steps_per_checkpoint) / steps_per_checkpoint))
        assert np.isclose(curr_lr, expected_lr), f"step: {step}"

        if prev_lr < curr_lr:
            # Since Cosine decay is periodic, we should see the learning rate
            # decrease and then increase again.
            num_restarts += 1

        prev_lr = curr_lr
        scheduler.step()

        curr_lr = optimizer.param_groups[0]["lr"]

    assert num_restarts == 10, f"num_restarts: {num_restarts}"


def test_lr_scheduler_cosine_decay_increasing_period():
    total_steps = 20000
    steps_per_checkpoint = 1000
    base_lr = 1.0

    module = NumberInputFeature(NumberInputFeatureConfig(name="num1", encoder=DenseEncoderConfig()))

    optimizer = SGD(module.parameters(), lr=base_lr)
    config = LRSchedulerConfig(
        decay="cosine",
        t_0=steps_per_checkpoint,
        t_mult=2,
        decay_rate=0,
        reduce_on_plateau=0,
    )
    scheduler = LRScheduler(
        config=config, optimizer=optimizer, steps_per_checkpoint=steps_per_checkpoint, total_steps=total_steps
    )

    curr_lr = base_lr
    prev_lr = base_lr
    num_restarts = 0
    for _ in range(total_steps + 1):
        if prev_lr < curr_lr:
            # Since Cosine decay is periodic, we should see the learning rate
            # decrease and then increase again.
            num_restarts += 1

        prev_lr = curr_lr
        scheduler.step()

        curr_lr = optimizer.param_groups[0]["lr"]

    # 1000, 3000, 6000, 12000, 24000 (but we stop at 20000)
    assert num_restarts == 4, f"num_restarts: {num_restarts}"


def test_lr_scheduler_save_load():
    steps_per_checkpoint = 10
    total_steps = 100
    base_lr = 1.0
    reduce_limit = 3

    module = NumberInputFeature(NumberInputFeatureConfig(name="num1", encoder=DenseEncoderConfig()))
    output1 = NumberOutputFeature(ECDNumberOutputFeatureConfig(name="output1", input_size=10), output_features={})

    optimizer = SGD(module.parameters(), lr=base_lr)
    config = LRSchedulerConfig(warmup_fraction=0.2, reduce_on_plateau=reduce_limit)
    scheduler = LRScheduler(
        config=config, optimizer=optimizer, steps_per_checkpoint=steps_per_checkpoint, total_steps=total_steps
    )

    progress_tracker = get_new_progress_tracker(
        batch_size=64,
        best_eval_metric_value=float("inf"),
        best_increase_batch_size_eval_metric=float("inf"),
        learning_rate=base_lr,
        output_features={"output1": output1},
    )

    for _ in range(10):
        scheduler.step()

    progress_tracker.train_metrics["output1"]["loss"].append(TrainerMetric(epoch=0, step=10, value=1.0))
    scheduler.eval_step(progress_tracker, "output1")

    optimizer_state = optimizer.state_dict()
    scheduler_state = scheduler.state_dict()

    optimizer2 = SGD(module.parameters(), lr=base_lr)
    scheduler2 = LRScheduler(
        config=config, optimizer=optimizer2, steps_per_checkpoint=steps_per_checkpoint, total_steps=total_steps
    )

    # Important: state needs to be loaded after init of optimizer and scheduler, otherwise
    # it can override loaded state
    optimizer2.load_state_dict(optimizer_state)
    scheduler2.load_state_dict(scheduler_state)

    lr = optimizer.param_groups[0]["lr"]
    assert lr == optimizer2.param_groups[0]["lr"]
    assert scheduler.state_dict() == scheduler2.state_dict()

    for _ in range(10):
        scheduler.step()
        scheduler2.step()

    progress_tracker.train_metrics["output1"]["loss"].append(TrainerMetric(epoch=1, step=20, value=0.8))
    scheduler.eval_step(progress_tracker, "output1")
    scheduler2.eval_step(progress_tracker, "output1")

    assert lr != optimizer.param_groups[0]["lr"]
    assert optimizer.param_groups[0]["lr"] == optimizer2.param_groups[0]["lr"]
    assert scheduler.state_dict() == scheduler2.state_dict()



import logging
from collections import defaultdict
from typing import Dict, List, Tuple, TYPE_CHECKING

try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import AUTO, COMBINED, LOSS
from ludwig.models.base import BaseModel
from ludwig.modules.metric_modules import get_best_function
from ludwig.utils.data_utils import save_json
from ludwig.utils.metric_utils import TrainerMetric

if TYPE_CHECKING:
    from ludwig.features.base_feature import OutputFeature
    from ludwig.schema.trainer import BaseTrainerConfig

logger = logging.getLogger(__name__)


@DeveloperAPI
def initialize_trainer_metric_dict(output_features) -> Dict[str, Dict[str, List[TrainerMetric]]]:
    """Returns a dict of dict of metrics, output_feature_name -> metric_name -> List[TrainerMetric]."""
    metrics = defaultdict(lambda: defaultdict(list))
    return metrics


def get_latest_metrics_dict(
    progress_tracker_metrics: Dict[str, Dict[str, List[TrainerMetric]]]
) -> Dict[str, Dict[str, float]]:
    """Returns a dict of field name -> metric name -> latest metric value."""
    latest_metrics_dict = defaultdict(dict)
    for feature_name, metrics_dict in progress_tracker_metrics.items():
        for metric_name, metrics in metrics_dict.items():
            if metrics:
                # Metrics may be missing if computing metrics was excepted, if the metrics are entirely empty
                # due to a missing subset, or if evaluate_training_set is False.
                latest_metrics_dict[feature_name][metric_name] = metrics[-1][-1]
    return latest_metrics_dict


@DeveloperAPI
def get_new_progress_tracker(
    batch_size: int,
    best_eval_metric_value: float,
    best_increase_batch_size_eval_metric: float,
    learning_rate: float,
    output_features: Dict[str, "OutputFeature"],
):
    """Returns a new instance of a ProgressTracker with empty metrics."""
    return ProgressTracker(
        epoch=0,
        batch_size=batch_size,
        steps=0,
        tune_checkpoint_num=0,
        checkpoint_number=0,
        best_eval_metric_steps=0,
        best_eval_metric_epoch=0,
        best_eval_metric_checkpoint_number=0,
        last_learning_rate_reduction_steps=0,
        last_increase_batch_size_steps=0,
        last_improvement_steps=0,
        best_eval_metric_value=best_eval_metric_value,
        best_increase_batch_size_eval_metric=best_increase_batch_size_eval_metric,
        last_increase_batch_size_eval_metric_improvement=0,
        learning_rate=learning_rate,
        num_reductions_learning_rate=0,
        num_increases_batch_size=0,
        train_metrics=initialize_trainer_metric_dict(output_features),
        validation_metrics=initialize_trainer_metric_dict(output_features),
        test_metrics=initialize_trainer_metric_dict(output_features),
        last_learning_rate_reduction=0,
        last_increase_batch_size=0,
        best_eval_train_metrics={},
        best_eval_validation_metrics={},
        best_eval_test_metrics={},
    )


@DeveloperAPI
class ProgressTracker:
    def __init__(
        self,
        epoch: int,
        batch_size: int,
        steps: int,
        tune_checkpoint_num: int,
        checkpoint_number: int,
        best_eval_metric_steps: int,
        best_eval_metric_epoch: int,
        best_eval_metric_checkpoint_number: int,
        last_improvement_steps: int,
        last_learning_rate_reduction_steps: int,
        last_increase_batch_size_steps: int,
        best_eval_metric_value: float,
        best_increase_batch_size_eval_metric: float,
        last_increase_batch_size_eval_metric_improvement: int,
        learning_rate: float,
        num_reductions_learning_rate: int,
        num_increases_batch_size: int,
        train_metrics: Dict[str, Dict[str, List[TrainerMetric]]],
        validation_metrics: Dict[str, Dict[str, List[TrainerMetric]]],
        test_metrics: Dict[str, Dict[str, List[TrainerMetric]]],
        last_learning_rate_reduction: int,
        last_increase_batch_size: int,
        best_eval_train_metrics: Dict[str, Dict[str, float]],
        best_eval_validation_metrics: Dict[str, Dict[str, float]],
        best_eval_test_metrics: Dict[str, Dict[str, float]],
    ):
        """JSON-serializable holder object that stores information related to training progress.

        [train/vali/test]_metrics is a nested dictionary of TrainerMetrics: feature_name -> metric_name ->
        List[TrainerMetrics], with one entry per training checkpoint.

        Note that when a model resumes training from a checkpoint, the progress tracker is deserialized from JSON, which
        automatically converts TrainerMetrics namedtuples into regular (epoch, steps, value) tuples.

        Args:
            epoch: The current epoch number.
            steps: The current step of training.
            batch_size: The current batch size.
            tune_checkpoint_num: The hyperopt checkpoint number (Ray Tune).
            checkpoint_number: The current checkpoint number.

            best_eval_metric_steps: The step of training that has the best evaluation so far.
            best_eval_metric_epoch: The epoch of training that has the best evaluation so far.
            best_eval_metric_checkpoint_number: The checkpoint number that has the best evaluation so far.

            last_improvement_steps: The number of steps since the last improvement.
            last_learning_rate_reduction_steps: The training step of the last learning rate reduction.
            last_increase_batch_size_steps: The training_step of the the last batch size increase.

            best_eval_metric_value: The metric value of the best evaluation so far.
            best_increase_batch_size_eval_metric:
                The metric value of the best evaluation so far, for increasing the batch size.

            last_learning_rate_reduction: The number of steps since the last learning rate reduction.
            last_increase_batch_size: The number of steps since the last batch size increase.

            last_increase_batch_size_eval_metric_improvement:
                The number of checkpoints since the last batch size increase.

            num_reductions_learning_rate: The number of total reductions in learning rate.
            num_increases_batch_size: The number of total increases in batch size.

            train_metrics: Training metrics. <output feature name> -> <metric name> -> History of metrics.
            validation_metrics: Validation metrics. <output feature name> -> <metric name> -> History of metrics.
            test_metrics: Test metrics. <output feature name> -> <metric name> -> History of metrics.

            best_eval_train_metrics:
                Best eval train metrics: <output feature name> -> <metric name> -> <metric value>.
            best_eval_validation_metrics:
                Best eval validation metrics: <output feature name> -> <metric name> -> <metric value>.
            best_eval_test_metrics:
                Best eval test metrics: <output feature name> -> <metric name> -> <metric value>.
        """
        self.batch_size = batch_size
        self.epoch = epoch
        self.steps = steps
        self.tune_checkpoint_num = tune_checkpoint_num
        self.checkpoint_number = checkpoint_number
        self.best_eval_metric_steps = best_eval_metric_steps
        self.best_eval_metric_epoch = best_eval_metric_epoch
        self.best_eval_metric_checkpoint_number = best_eval_metric_checkpoint_number
        self.last_improvement_steps = last_improvement_steps
        self.last_learning_rate_reduction_steps = last_learning_rate_reduction_steps
        self.last_learning_rate_reduction = last_learning_rate_reduction
        self.last_increase_batch_size_steps = last_increase_batch_size_steps
        self.last_increase_batch_size = last_increase_batch_size
        self.learning_rate = learning_rate
        self.best_eval_metric_value = best_eval_metric_value
        self.best_increase_batch_size_eval_metric = best_increase_batch_size_eval_metric
        self.last_increase_batch_size_eval_metric_improvement = last_increase_batch_size_eval_metric_improvement
        self.num_reductions_learning_rate = num_reductions_learning_rate
        self.num_increases_batch_size = num_increases_batch_size
        self.train_metrics = train_metrics
        self.validation_metrics = validation_metrics
        self.test_metrics = test_metrics

        # Best metrics.
        self.best_eval_train_metrics = best_eval_train_metrics
        self.best_eval_validation_metrics = best_eval_validation_metrics
        self.best_eval_test_metrics = best_eval_test_metrics

    def save(self, filepath):
        save_json(filepath, self.__dict__)

    @staticmethod
    def load(progress_tracking_dict: Dict):
        from ludwig.utils.backward_compatibility import upgrade_model_progress

        loaded = upgrade_model_progress(progress_tracking_dict)
        return ProgressTracker(**loaded)

    def log_metrics(self):
        log_metrics = {
            "batch_size": self.batch_size,
            "epoch": self.epoch,
            "steps": self.steps,
            "tune_checkpoint_num": self.tune_checkpoint_num,
            "checkpoint_number": self.checkpoint_number,
            "last_improvement_steps": self.last_improvement_steps,
            "best_eval_metric_steps": self.best_eval_metric_steps,
            "best_eval_metric_epoch": self.best_eval_metric_epoch,
            "best_eval_metric_checkpoint_number": self.best_eval_metric_checkpoint_number,
            "learning_rate": self.learning_rate,
            "best_valid_metric": self.best_eval_metric_value,
            "num_reductions_lr": self.num_reductions_learning_rate,
            "num_increases_bs": self.num_increases_batch_size,
        }
        for metrics_dict_name in [
            "train_metrics",
            "validation_metrics",
            "test_metrics",
        ]:
            metrics_dict = getattr(self, metrics_dict_name)
            for feature_name in metrics_dict:
                for metric_name, metrics_tuples in metrics_dict[feature_name].items():
                    if metrics_tuples:
                        # For logging, get the latest metrics. The second "-1" indexes into the TrainerMetric
                        # namedtuple. The last element of the TrainerMetric namedtuple is the actual metric value.
                        #
                        # TODO: when loading an existing model, this loses metric values for all but the last epoch.
                        log_metrics[f"{metrics_dict_name}.{feature_name}.{metric_name}"] = metrics_tuples[-1][-1]

        # Add best metrics.
        for feature_name, metrics in self.best_eval_train_metrics.items():
            for metric_name, metric_value in metrics.items():
                log_metrics[f"best.train_metrics.{feature_name}.{metric_name}"] = metric_value
        for feature_name, metrics in self.best_eval_validation_metrics.items():
            for metric_name, metric_value in metrics.items():
                log_metrics[f"best.validation_metrics.{feature_name}.{metric_name}"] = metric_value
        for feature_name, metrics in self.best_eval_test_metrics.items():
            for metric_name, metric_value in metrics.items():
                log_metrics[f"best.test_metrics.{feature_name}.{metric_name}"] = metric_value

        return log_metrics


@DeveloperAPI
def append_metrics(
    model: BaseModel,
    dataset_name: Literal["train", "validation", "test"],
    results: Dict[str, Dict[str, float]],
    metrics_log: Dict[str, Dict[str, List[TrainerMetric]]],
    progress_tracker: ProgressTracker,
) -> Dict[str, Dict[str, List[TrainerMetric]]]:
    epoch = progress_tracker.epoch
    steps = progress_tracker.steps
    for output_feature in model.output_features:
        scores = [dataset_name]

        # collect metric names based on output features metrics to
        # ensure consistent order of reporting metrics
        metric_names = sorted(results[output_feature].keys())

        for metric in metric_names:
            if metric in results[output_feature]:
                # Some metrics may have been excepted and excluded from results.
                score = results[output_feature][metric]
                metrics_log[output_feature][metric].append(TrainerMetric(epoch=epoch, step=steps, value=score))
                scores.append(score)

    metrics_log[COMBINED][LOSS].append(TrainerMetric(epoch=epoch, step=steps, value=results[COMBINED][LOSS]))
    return metrics_log


@DeveloperAPI
def get_total_steps(epochs: int, steps_per_epoch: int, train_steps: int):
    """Returns train_steps if provided, otherwise epochs * steps_per_epoch."""
    if train_steps:
        return train_steps
    return epochs * steps_per_epoch


@DeveloperAPI
def get_final_steps_per_checkpoint(
    steps_per_epoch: int, steps_per_checkpoint: int = 0, checkpoints_per_epoch: float = 0, should_log: bool = False
):
    """Returns the steps per checkpoint to use for the training loop, given user+default inputs."""
    if steps_per_checkpoint != 0 and checkpoints_per_epoch != 0:
        raise ValueError(
            "It is invalid to specify both checkpoints_per_epoch AND steps_per_checkpoint. Please specify one or the "
            "other, or specify neither to checkpoint/eval the model every epoch."
        )

    # Set steps_per_checkpoint based on the checkpoints_per_epoch, if checkpoints_per_epoch was specified.
    if checkpoints_per_epoch != 0:
        steps_per_checkpoint = int(steps_per_epoch / checkpoints_per_epoch)

    # Cap steps_per_checkpoint at steps_per_epoch.
    if steps_per_checkpoint > steps_per_epoch:
        if should_log:
            logger.info(
                f"Note: steps_per_checkpoint (was {steps_per_checkpoint}) is now set to the number of "
                f"steps per epoch: {steps_per_epoch}.\n"
            )
        return steps_per_epoch

    # steps_per_checkpoint wasn't specified. Use steps_per_epoch.
    if steps_per_checkpoint == 0:
        return steps_per_epoch

    return steps_per_checkpoint


@DeveloperAPI
def get_training_report(
    validation_field: str,
    validation_metric: str,
    include_test_set: bool,
    train_valiset_stats: Dict[str, Dict[str, List[float]]],
    train_testset_stats: Dict[str, Dict[str, List[float]]],
) -> List[Tuple[str, str]]:
    """Returns a training report in the form of a list [(report item, value)]."""
    validation_field_result = train_valiset_stats[validation_field]
    best_function = get_best_function(validation_metric)

    training_report = []
    best_vali_index, (
        epoch_best_validation_metric,
        step_best_validation_metric,
        best_validation_metric,
    ) = best_function(
        enumerate(validation_field_result[validation_metric]),
        # -1 for the last element of the TrainerMetric namedtuple.
        key=lambda index_epoch_step_value: index_epoch_step_value[1][-1],
    )
    training_report.append(["Validation feature", validation_field])
    training_report.append(["Validation metric", validation_metric])
    training_report.append(["Best model step", step_best_validation_metric])
    training_report.append(["Best model epoch", epoch_best_validation_metric + 1])
    training_report.append(
        [
            f"Best model's validation {validation_metric}",
            best_validation_metric,
        ]
    )
    if include_test_set:
        validation_selected_test_metric_score = train_testset_stats[validation_field][validation_metric][
            best_vali_index
        ][
            -1
        ]  # -1 for the last element of the TrainerMetric namedtuple.

        training_report.append(
            [
                f"Best model's test {validation_metric}",
                validation_selected_test_metric_score,
            ]
        )
    return training_report


def get_rendered_batch_size_grad_accum(config: "BaseTrainerConfig", num_workers: int) -> Tuple[int, int]:
    """Returns the batch size and gradient accumulation steps to use for training.

    For batch_size==AUTO:
    1. effective_batch_size is not AUTO and gradient_accumulation_steps is not AUTO:
        batch size is set to the effective batch size divided by the gradient accumulation steps, divided by the
        number of workers.
    2. effective_batch_size is AUTO or gradient_accumulation_steps is AUTO:
        batch size remains AUTO.

    For gradient_accumulation_steps==AUTO:
    1. batch size is AUTO:
        gradient accumulation steps remains AUTO.
    2. batch_size is not AUTO and effective batch size is not AUTO:
        gradient accumulation steps is set to the effective batch size divided by the batch size, divided by the number
        of workers.
    3. batch size is not AUTO and effective batch size is AUTO:
        gradient accumulation steps is set to 1.
    """
    effective_batch_size = config.effective_batch_size
    batch_size = config.batch_size
    gradient_accumulation_steps = config.gradient_accumulation_steps

    if config.batch_size == AUTO:
        if config.effective_batch_size != AUTO and config.gradient_accumulation_steps != AUTO:
            batch_size = max(int(effective_batch_size / gradient_accumulation_steps / num_workers), 1)

    if config.gradient_accumulation_steps == AUTO:
        if config.batch_size != AUTO:
            if config.effective_batch_size != AUTO:
                gradient_accumulation_steps = max(int(effective_batch_size / batch_size / num_workers), 1)
            else:
                gradient_accumulation_steps = 1

    return batch_size, gradient_accumulation_steps



from typing import Optional

from ludwig.api_annotations import DeveloperAPI
from ludwig.schema import utils as schema_utils
from ludwig.schema.defaults.llm import LLMDefaultsConfig, LLMDefaultsField
from ludwig.schema.features.base import (
    BaseInputFeatureConfig,
    BaseOutputFeatureConfig,
    FeatureCollection,
    LLMInputFeatureSelection,
    LLMOutputFeatureSelection,
)
from ludwig.schema.hyperopt import HyperoptConfig, HyperoptField
from ludwig.schema.llms.base_model import BaseModelDataclassField
from ludwig.schema.llms.generation import LLMGenerationConfig, LLMGenerationConfigField
from ludwig.schema.llms.model_parameters import ModelParametersConfig, ModelParametersConfigField
from ludwig.schema.llms.peft import AdapterDataclassField, BaseAdapterConfig
from ludwig.schema.llms.prompt import PromptConfig, PromptConfigField
from ludwig.schema.llms.quantization import QuantizationConfig, QuantizationConfigField
from ludwig.schema.model_types.base import ModelConfig, register_model_type
from ludwig.schema.preprocessing import PreprocessingConfig, PreprocessingField
from ludwig.schema.trainer import LLMTrainerConfig, LLMTrainerDataclassField
from ludwig.schema.utils import ludwig_dataclass


@DeveloperAPI
@register_model_type(name="llm")
@ludwig_dataclass
class LLMModelConfig(ModelConfig):
    """Parameters for LLM Model Type."""

    model_type: str = schema_utils.ProtectedString("llm")

    base_model: str = BaseModelDataclassField()

    input_features: FeatureCollection[BaseInputFeatureConfig] = LLMInputFeatureSelection().get_list_field()
    output_features: FeatureCollection[BaseOutputFeatureConfig] = LLMOutputFeatureSelection().get_list_field()

    preprocessing: PreprocessingConfig = PreprocessingField().get_default_field()
    defaults: Optional[LLMDefaultsConfig] = LLMDefaultsField().get_default_field()
    hyperopt: Optional[HyperoptConfig] = HyperoptField().get_default_field()

    prompt: PromptConfig = PromptConfigField().get_default_field()

    # trainer: LLMTrainerConfig = LLMTrainerField().get_default_field()
    trainer: LLMTrainerConfig = LLMTrainerDataclassField(
        description="The trainer to use for the model",
    )

    generation: LLMGenerationConfig = LLMGenerationConfigField().get_default_field()

    adapter: Optional[BaseAdapterConfig] = AdapterDataclassField()
    quantization: Optional[QuantizationConfig] = QuantizationConfigField().get_default_field()
    model_parameters: Optional[ModelParametersConfig] = ModelParametersConfigField().get_default_field()



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import functools
import json
import logging

import numpy as np
import pandas as pd
import pytest
from fsspec.config import conf

from ludwig.api import LudwigModel
from ludwig.data.cache.types import CacheableDataframe
from ludwig.data.dataset_synthesizer import build_synthetic_dataset_df
from ludwig.utils.data_utils import (
    add_sequence_feature_column,
    figure_data_format_dataset,
    get_abs_path,
    hash_dict,
    NumpyEncoder,
    PANDAS_DF,
    read_csv,
    read_html,
    read_parquet,
    sanitize_column_names,
    use_credentials,
)
from tests.integration_tests.utils import private_param

try:
    import dask.dataframe as dd
except ImportError:
    dd = None


def test_add_sequence_feature_column():
    df = pd.DataFrame([1, 2, 3, 4, 5], columns=["x"])

    add_sequence_feature_column(df, "x", 2)
    assert df.equals(
        pd.DataFrame(
            [
                [1, "1 2"],
                [2, "1 2"],
                [3, "1 2"],
                [4, "2 3"],
                [5, "3 4"],
            ],
            columns=["x", "x_feature"],
        )
    )

    add_sequence_feature_column(df, "x", 1)
    assert df.equals(
        pd.DataFrame(
            [
                [1, "1"],
                [2, "1"],
                [3, "2"],
                [4, "3"],
                [5, "4"],
            ],
            columns=["x", "x_feature"],
        )
    )

    df = pd.DataFrame([1, 2, 3, 4, 5], columns=["x"])

    add_sequence_feature_column(df, "y", 2)
    assert df.equals(pd.DataFrame([1, 2, 3, 4, 5], columns=["x"]))


def test_get_abs_path():
    assert get_abs_path("a", "b.jpg") == "a/b.jpg"
    assert get_abs_path(None, "b.jpg") == "b.jpg"


@pytest.mark.parametrize(
    "path, expected_format", [("s3://path/to.parquet ", "parquet"), ("/Users/path/to.csv \n", "csv")]
)
def test_figure_data_format_dataset_strip(path, expected_format):
    assert figure_data_format_dataset(path) == expected_format


@pytest.mark.distributed
def test_figure_data_format_dataset():
    assert figure_data_format_dataset({"a": "b"}) == dict
    assert figure_data_format_dataset(pd.DataFrame([1, 2, 3, 4, 5], columns=["x"])) == pd.DataFrame
    assert (
        figure_data_format_dataset(
            dd.from_pandas(pd.DataFrame([1, 2, 3, 4, 5], columns=["x"]), npartitions=1).reset_index()
        )
        == dd.core.DataFrame
    )
    assert (
        figure_data_format_dataset(
            CacheableDataframe(df=pd.DataFrame([1, 2, 3, 4, 5], columns=["x"]), name="test", checksum="test123")
        )
        == pd.DataFrame
    )
    assert (
        figure_data_format_dataset(
            CacheableDataframe(
                df=dd.from_pandas(pd.DataFrame([1, 2, 3, 4, 5], columns=["x"]), npartitions=1).reset_index(),
                name="test",
                checksum="test123",
            )
        )
        == dd.core.DataFrame
    )


def test_hash_dict_numpy_types():
    d = {"float32": np.float32(1)}
    assert hash_dict(d) == b"uqtgWB"


def test_use_credentials():
    conf.clear()
    with use_credentials(None):
        assert len(conf) == 0

    s3_creds = {
        "s3": {
            "client_kwargs": {
                "endpoint_url": "http://localhost:9000",
                "aws_access_key_id": "test",
                "aws_secret_access_key": "test",
            }
        }
    }
    with use_credentials(s3_creds):
        assert len(conf) == 1
        assert conf == s3_creds

    assert len(conf) == 0


def test_numpy_encoder():
    # Test Python builtin data type encoding.
    assert json.dumps(None, cls=NumpyEncoder) == "null"
    assert json.dumps({}, cls=NumpyEncoder) == "{}"
    assert json.dumps(1, cls=NumpyEncoder) == "1"
    assert json.dumps(1.0, cls=NumpyEncoder) == "1.0"
    assert json.dumps("a", cls=NumpyEncoder) == '"a"'
    assert json.dumps([0, 1, 2, 3, 4], cls=NumpyEncoder) == "[0, 1, 2, 3, 4]"
    assert json.dumps((0, 1, 2, 3, 4), cls=NumpyEncoder) == "[0, 1, 2, 3, 4]"
    assert json.dumps({0, 1, 2, 3, 4}, cls=NumpyEncoder) == "[0, 1, 2, 3, 4]"
    assert json.dumps({"a": "b"}, cls=NumpyEncoder) == '{"a": "b"}'

    # Test numpy data type encoding
    for dtype in [np.byte, np.ubyte, np.short, np.ushort, np.int32, np.int64, np.uint, np.longlong, np.ulonglong]:
        x = np.arange(5, dtype=dtype)
        assert json.dumps(x, cls=NumpyEncoder) == "[0, 1, 2, 3, 4]"
        for i in x:
            assert json.dumps(i, cls=NumpyEncoder) == f"{i}"

    for dtype in [np.half, np.single, np.double, np.longdouble]:
        x = np.arange(5, dtype=dtype)
        assert json.dumps(x, cls=NumpyEncoder) == "[0.0, 1.0, 2.0, 3.0, 4.0]"
        for i in x:
            assert json.dumps(i, cls=NumpyEncoder) == f"{i}"


def test_dataset_synthesizer_output_feature_decoder():
    config = {
        "input_features": [{"name": "sentence", "type": "text"}],
        "output_features": [{"name": "product", "type": "category"}],
        "trainer": {"epochs": 5},
        "model_type": "ecd",
    }
    build_synthetic_dataset_df(dataset_size=100, config=config)
    LudwigModel(config=config, logging_level=logging.INFO)


@pytest.mark.parametrize(
    "dataset_1k_url",
    [
        private_param(["s3://ludwig-tests/datasets/synthetic_1k.csv"]),
        private_param(["s3://ludwig-tests/datasets/synthetic_1k.parquet"]),
    ],
)
@pytest.mark.parametrize("nrows", [None, 100])
def test_chunking(dataset_1k_url, nrows):
    reader_fn = {"csv": read_csv, "parquet": functools.partial(read_parquet, df_lib=PANDAS_DF)}

    format = figure_data_format_dataset(dataset_1k_url)

    assert reader_fn[format](dataset_1k_url, nrows=nrows).shape[0] == (nrows if nrows else 1000)


@pytest.mark.parametrize(
    "df_lib", [pytest.param(pd, id="pandas"), pytest.param(dd, marks=pytest.mark.distributed, id="dask")]
)
@pytest.mark.parametrize("nrows", [None, 10])
def test_read_html(df_lib, nrows):
    HTML_DOCUMENT = """
    <!DOCTYPE html>
    <html>
    <head><title>TITLE</title></head>
    <body>
    <table>
    <th><td>Col 1</td><td>Col 2</td></th>
    <tr><td>1</td><td>2</td></tt>
    </table>
    </body>
    </html>
    """

    kwargs = {}
    if not nrows:
        kwargs["nrows"] = nrows

    read_html(HTML_DOCUMENT, df_lib, **kwargs)


def test_sanitize_column_names():
    df = pd.DataFrame(
        {
            "col.one": [1, 2, 3, 4],
            "col(two)": [4, 5, 6, 7],
            "col[]:three": [7, 8, 9, 10],
            "col 'one' (new)": [1, 2, 3, 4],
        }
    )

    df = sanitize_column_names(df)

    assert list(df.columns) == ["col_one", "col_two_", "col___three", "col _one_ _new_"]



from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import RANDOM
from ludwig.schema import utils as schema_utils
from ludwig.schema.metadata import PREPROCESSING_METADATA
from ludwig.schema.split import BaseSplitConfig, SplitDataclassField
from ludwig.schema.utils import ludwig_dataclass


@DeveloperAPI
@ludwig_dataclass
class PreprocessingConfig(schema_utils.BaseMarshmallowConfig):
    """Global preprocessing config is a dataclass that configures the parameters used for global preprocessing."""

    sample_ratio: float = schema_utils.NonNegativeFloat(
        default=1.0,
        description="The ratio of the dataset to use. For instance, if 0.5, half of the dataset "
        "provided will be used.",
        parameter_metadata=PREPROCESSING_METADATA["sample_ratio"],
    )

    oversample_minority: float = schema_utils.NonNegativeFloat(
        default=None,
        allow_none=True,
        description="If not None, the minority class will be oversampled to reach the specified ratio respective to "
        "the majority class. ",
        parameter_metadata=PREPROCESSING_METADATA["oversample_minority"],
    )

    undersample_majority: float = schema_utils.NonNegativeFloat(
        default=None,
        allow_none=True,
        description="If not None, the majority class will be undersampled to reach the specified ratio respective "
        "to the minority class. ",
        parameter_metadata=PREPROCESSING_METADATA["undersample_majority"],
    )

    split: BaseSplitConfig = SplitDataclassField(
        default=RANDOM,
    )

    global_max_sequence_length: int = schema_utils.PositiveInteger(
        default=None,
        allow_none=True,
        description="Specifically for LLMs. This is the maximum length of the input sequence going into the model's "
        "forward pass during training. Sequences will be truncated to this length after merging inputs and targets. "
        "If not set, the total length of the merged input and target token sequences will be used.",
        parameter_metadata=PREPROCESSING_METADATA["global_max_sequence_length"],
    )


@DeveloperAPI
class PreprocessingField(schema_utils.DictMarshmallowField):
    def __init__(self):
        super().__init__(PreprocessingConfig)



from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import BINARY, CATEGORY, NUMBER
from ludwig.schema import utils as schema_utils
from ludwig.schema.defaults.base import BaseDefaultsConfig
from ludwig.schema.defaults.utils import DefaultsDataclassField
from ludwig.schema.features.base import BaseFeatureConfig
from ludwig.schema.features.utils import gbm_defaults_config_registry
from ludwig.schema.utils import ludwig_dataclass


@DeveloperAPI
@ludwig_dataclass
class GBMDefaultsConfig(BaseDefaultsConfig):
    binary: BaseFeatureConfig = DefaultsDataclassField(
        feature_type=BINARY, defaults_registry=gbm_defaults_config_registry
    )

    category: BaseFeatureConfig = DefaultsDataclassField(
        feature_type=CATEGORY, defaults_registry=gbm_defaults_config_registry
    )

    number: BaseFeatureConfig = DefaultsDataclassField(
        feature_type=NUMBER, defaults_registry=gbm_defaults_config_registry
    )


@DeveloperAPI
class GBMDefaultsField(schema_utils.DictMarshmallowField):
    def __init__(self):
        super().__init__(GBMDefaultsConfig)



from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import TEXT
from ludwig.schema import utils as schema_utils
from ludwig.schema.defaults.base import BaseDefaultsConfig
from ludwig.schema.defaults.utils import DefaultsDataclassField
from ludwig.schema.features.base import BaseFeatureConfig
from ludwig.schema.features.utils import llm_defaults_config_registry
from ludwig.schema.utils import ludwig_dataclass


@DeveloperAPI
@ludwig_dataclass
class LLMDefaultsConfig(BaseDefaultsConfig):
    text: BaseFeatureConfig = DefaultsDataclassField(feature_type=TEXT, defaults_registry=llm_defaults_config_registry)


@DeveloperAPI
class LLMDefaultsField(schema_utils.DictMarshmallowField):
    def __init__(self):
        super().__init__(LLMDefaultsConfig)



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from typing import Any, Dict, List, Optional, Type

import torch

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import BAG, ENCODER_OUTPUT
from ludwig.encoders.base import Encoder
from ludwig.encoders.registry import register_encoder
from ludwig.encoders.types import EncoderOutputDict
from ludwig.modules.embedding_modules import EmbedWeighted
from ludwig.modules.fully_connected_modules import FCStack
from ludwig.schema.encoders.bag_encoders import BagEmbedWeightedConfig
from ludwig.schema.encoders.base import BaseEncoderConfig

logger = logging.getLogger(__name__)


@DeveloperAPI
@register_encoder("embed", BAG)
class BagEmbedWeightedEncoder(Encoder):
    def __init__(
        self,
        vocab: List[str],
        embedding_size: int = 50,
        representation: str = "dense",
        embeddings_trainable: bool = True,
        pretrained_embeddings: Optional[str] = None,
        force_embedding_size: bool = False,
        embeddings_on_cpu: bool = False,
        fc_layers=None,
        num_fc_layers: int = 0,
        output_size: int = 10,
        use_bias: bool = True,
        weights_initializer: str = "xavier_uniform",
        bias_initializer: str = "zeros",
        norm: Optional[str] = None,
        norm_params: Optional[Dict[str, Any]] = None,
        activation: str = "relu",
        dropout: float = 0.0,
        encoder_config=None,
        **kwargs,
    ):
        super().__init__()
        self.config = encoder_config

        logger.debug(f" {self.name}")

        logger.debug("  EmbedWeighted")
        self.embed_weighted = EmbedWeighted(
            vocab,
            embedding_size,
            representation=representation,
            embeddings_trainable=embeddings_trainable,
            pretrained_embeddings=pretrained_embeddings,
            force_embedding_size=force_embedding_size,
            embeddings_on_cpu=embeddings_on_cpu,
            dropout=dropout,
            embedding_initializer=weights_initializer,
        )
        logger.debug("  FCStack")
        self.fc_stack = FCStack(
            self.embed_weighted.output_shape[-1],
            layers=fc_layers,
            num_layers=num_fc_layers,
            default_output_size=output_size,
            default_use_bias=use_bias,
            default_weights_initializer=weights_initializer,
            default_bias_initializer=bias_initializer,
            default_norm=norm,
            default_norm_params=norm_params,
            default_activation=activation,
            default_dropout=dropout,
        )

    @staticmethod
    def get_schema_cls() -> Type[BaseEncoderConfig]:
        return BagEmbedWeightedConfig

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([len(self.vocab)])

    @property
    def output_shape(self) -> torch.Size:
        return self.fc_stack.output_shape

    def forward(self, inputs: torch.Tensor) -> EncoderOutputDict:
        """
        :param inputs: The inputs fed into the encoder.
               Shape: [batch x vocab size], type torch.int32

        :param return: embeddings of shape [batch x embed size], type torch.float32
        """
        hidden = self.embed_weighted(inputs)
        hidden = self.fc_stack(hidden)

        return {ENCODER_OUTPUT: hidden}



import logging
from typing import Optional, Tuple, TYPE_CHECKING

import torch
from torch import nn
from torch.distributed.fsdp import FullyShardedDataParallel as FSDP
from torch.optim import Optimizer

from ludwig.distributed.ddp import DDPStrategy
from ludwig.modules.optimization_modules import create_optimizer

if TYPE_CHECKING:
    from ludwig.schema.trainer import ECDTrainerConfig


class FSDPStrategy(DDPStrategy):
    def _log_on_init(self):
        logging.info("Using FSDP strategy")

    def prepare(
        self,
        model: nn.Module,
        trainer_config: "ECDTrainerConfig",
        base_learning_rate: float,
    ) -> Tuple[nn.Module, Optimizer]:
        return FSDP(model), create_optimizer(model, trainer_config.optimizer, base_learning_rate)

    def to_device(self, model: nn.Module, device: Optional[torch.device] = None) -> nn.Module:
        return model

    @classmethod
    def is_model_parallel(cls) -> bool:
        return True



"""Implements similar functionality as tf.train.Checkpoint and tf.train.CheckpointManager.

https://gist.github.com/kevinzakka/5d345421f7abefd5dbaf6a77f829e70a.
"""

import errno
import logging
import os
import re
import shutil
import signal
import tempfile
import uuid
from abc import ABC, abstractmethod
from glob import glob
from typing import Any, Dict, Mapping, Optional, TYPE_CHECKING

import torch
from torch.optim import Optimizer

from ludwig.api_annotations import DeveloperAPI
from ludwig.modules.lr_scheduler import LRScheduler

if TYPE_CHECKING:
    from ludwig.distributed.base import DistributedStrategy
    from ludwig.models.base import BaseModel


logger = logging.getLogger(__name__)
LATEST = "latest"
BEST = "best"


@DeveloperAPI
def mkdir(s):
    """Create a directory if it doesn't already exist."""
    if not os.path.exists(s):
        os.makedirs(s)


@DeveloperAPI
def get_files(d, pattern, sort=True):
    """Return a list of files in a given directory.

    Args:
      d (str): The path to the directory.
      pattern (str): The wildcard to filter files with.
      sort (bool): Whether to sort the returned list. Assumes filenames contain a number value to sort by (tmp-001).
    """
    files = glob(os.path.join(d, pattern))
    files = [f for f in files if os.path.isfile(f)]
    if sort:

        def filter_numeric(s):
            return re.sub("[^0-9]", "", s)

        files.sort(key=lambda x: int(filter_numeric(os.path.basename(x).split(".")[0])))
    return files


@DeveloperAPI
def get_latest_checkpoint_path(directory: str) -> str:
    latest_path = os.path.join(directory, f"{LATEST}.ckpt")
    if os.path.exists(latest_path):
        return latest_path

    # Legacy codepath for checkpoints saved by global step number
    ckpts = get_files(directory, "*.ckpt")
    if ckpts:
        return ckpts[-1]

    return None


@DeveloperAPI
class Checkpoint(ABC):
    """Save and restore model and optimizer states."""

    def __init__(
        self,
        distributed: "DistributedStrategy",
        model: "BaseModel",
        optimizer: Optional[Optimizer] = None,
        scheduler: Optional[LRScheduler] = None,
    ):
        """Constructor."""
        self.distributed = distributed
        self.model = model
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.global_step = 0

    def prepare(self, directory: str):
        # create checkpoint directory if it doesn't
        # already exist
        mkdir(directory)

    @abstractmethod
    def load(self, save_path: str, device: Optional[torch.device] = None) -> bool:
        pass

    @abstractmethod
    def get_state_for_inference(self, save_path: str, device: Optional[torch.device] = None) -> Mapping[str, Any]:
        pass

    @abstractmethod
    def save(self, save_path: str, global_step: int):
        pass

    def _get_global_step(self, state: Dict[str, Any], save_path: str) -> int:
        global_step = state.get("global_step")
        if global_step is None:
            # Legacy step detection for older checkpoint format which encoded the
            # step number in the checkpoint filename.
            return int(os.path.basename(save_path).split(".")[0])
        return global_step


@DeveloperAPI
class MultiNodeCheckpoint(Checkpoint):
    def prepare(self, directory: str):
        if self.is_local_rank_0():
            super().prepare(directory)
        self.distributed.barrier()

    def load(self, save_path: str, device: Optional[torch.device] = None) -> bool:
        """Load state from a saved checkpoint.

        Args:
          save_path (str): The filepath to the saved checkpoint.
          device (torch.device): The device on which to
            load the state.

        Returns:
          True if the checkpoint was sucessfully loaded, False if the checkpoint file
            could not be found.
        """
        try:
            state = torch.load(save_path, map_location=device)
            try:
                self.global_step = self._get_global_step(state, save_path)
                _, unexpected_keys = self.model.load_state_dict(state["model_weights"], strict=False)
                assert unexpected_keys == [], f"Unexpected keys found in state dict: {unexpected_keys}"
                if self.optimizer is not None:
                    self.optimizer.load_state_dict(state["optim_state"])
                if self.scheduler is not None and "scheduler_state" in state:
                    self.scheduler.load_state_dict(state["scheduler_state"])
                logger.info(f"Successfully loaded model weights from {save_path}.")
                return True
            except Exception as e:
                # there was an issue loading the state which means
                # either the model definition and saved weights
                # do not agree or they were not saved in the first
                # place.
                # since this is a severe issue, we raise an error
                # rather than allowing the program to proceed.
                raise e
        except FileNotFoundError as e:
            logger.error(e)
            return False

    def get_state_for_inference(self, save_path: str, device: Optional[torch.device] = None) -> Mapping[str, Any]:
        state = torch.load(save_path, map_location=device)
        return state["model_weights"]

    def save(self, save_path: str, global_step: int):
        """Save a state to disk.

        Modified from brentyi/fannypack.

        Args:
          save_path (str): The name of the checkpoint to save.
          global_step (int): The iteration number which will be used
             to name the checkpoint.
        """
        if self.is_local_rank_0():
            state = {
                "global_step": global_step,
                "model_weights": self.get_model_state_dict(),
            }
            if self.optimizer is not None:
                state["optim_state"] = self.optimizer.state_dict()
            if self.scheduler is not None:
                state["scheduler_state"] = self.scheduler.state_dict()

            # ignore ctrl+c while saving
            try:
                orig_handler = signal.getsignal(signal.SIGINT)
                signal.signal(signal.SIGINT, lambda _sig, _frame: None)
            except ValueError:
                # signal throws a ValueError if we're not in the main thread
                orig_handler = None

            try:
                # atomic save
                with tempfile.TemporaryDirectory() as tmpdir:
                    # Save to a temporary directory outside of the checkpoint dir so
                    # async processes do not try and copy a partially-written checkpoint.
                    # See Ray Tune and MLFlow for examples of background processes that
                    # are affected by this.
                    tmp_path = os.path.join(tmpdir, "temp.ckpt")
                    torch.save(state, tmp_path)

                    self.safe_move_file(tmp_path, save_path)
                    logger.debug(f"Saved checkpoint at {save_path}.")
            finally:
                # restore SIGINT handler
                if orig_handler is not None:
                    signal.signal(signal.SIGINT, orig_handler)
        self.distributed.barrier()

    def get_model_state_dict(self) -> Dict[str, Any]:
        state = self.model.state_dict()

        # Remove frozen parameter weights from state_dict for adapters and pretrained models
        for n, p in self.model.named_parameters():
            if not p.requires_grad:
                del state[n]

        return state

    def is_local_rank_0(self) -> bool:
        return self.distributed.local_rank() == 0

    def safe_move_file(self, src: str, dst: str):
        """Move a file from one directory to another, possibly across filesystems.

        This implementation specifically addresses the following issue with distributed training:

        1. The `save_path` is a directory local to the node, in which case every node should write
           checkpoints separately.
        2. The `save_path` is a remote / global filesystem like NFS, in which case only the coordinator
           should write checkpoints.
        """
        try:
            os.replace(src, dst)
        except OSError as err:
            if err.errno == errno.EXDEV:
                # Tried to move to an external filesystem. This means we should only run this on the coordinator
                if not self.distributed.is_coordinator():
                    logger.info(
                        f"Skipping writing checkpoint from rank {self.distributed.rank()} as it is not the coordinator "
                        f"and the destination filesystem is remote."
                    )
                    return

                # Generate a unique ID, and copy `<src>` to the target directory with a temporary name `<dst>.<ID>.tmp`.
                # Because we're copying across a filesystem boundary, this initial copy may not be atomic.  We insert a
                # random UUID so if different processes are copying into `<dst>`, they don't overlap in their tmp
                # copies.
                copy_id = uuid.uuid4()
                tmp_dst = f"{dst}.{copy_id}.tmp"
                shutil.copyfile(src, tmp_dst)

                # Atomic replace file onto the new name, and clean up original source file.
                os.replace(tmp_dst, dst)
                os.unlink(src)
            else:
                raise


@DeveloperAPI
class CheckpointManager:
    """A model and optimizer checkpoint manager."""

    def __init__(self, checkpoint: Checkpoint, directory: str, device: torch.device):
        """Constructor.

        Args:
          checkpoint (Checkpoint): An instance of `Checkpoint`.
          directory (str): The directory in which checkpoints will be saved.
          device (torch.device): The computing device on which to restore
            checkpoints.
        """
        self.checkpoint = checkpoint
        self.directory = directory
        self.device = device
        self.latest_checkpoint = None
        self.checkpoint.prepare(self.directory)

    def restore_or_initialize(self) -> int:
        """Restore items in checkpoint from the latest checkpoint file.

        Returns:
          The global iteration step. This is parsed from the latest
            checkpoint file if one is found, else 0 is returned.
        """
        last_ckpt = get_latest_checkpoint_path(self.directory)
        if last_ckpt:
            status = self.checkpoint.load(last_ckpt, self.device)
            if not status:
                logger.warning("Could not restore latest checkpoint file.")
                return 0
            self.latest_checkpoint = last_ckpt
            return self.checkpoint.global_step
        return 0

    def save(self, global_step: int, tag: str = LATEST):
        """Create a new checkpoint.

        Args:
           global_step (int): The iteration number which will be used
             to name the checkpoint.
        """
        save_path = os.path.join(self.directory, f"{tag}.ckpt")
        self.checkpoint.save(save_path, global_step)
        self.latest_checkpoint = save_path

    def save_best(self, global_step: int):
        self.save(global_step, BEST)

    def load(self, tag: str = LATEST):
        """Load a checkpoint.

        Args:
          tag (str): The tag of the checkpoint to load.
        """
        save_path = os.path.join(self.directory, f"{tag}.ckpt")
        self.checkpoint.load(save_path, self.device)

    def get_best_checkpoint_state_for_inference(self, device: torch.device) -> Mapping[str, Any]:
        save_path = os.path.join(self.directory, f"{BEST}.ckpt")
        return self.checkpoint.get_state_for_inference(save_path, device)

    def close(self):
        pass

    @staticmethod
    def load_latest_checkpoint(checkpoint: Checkpoint, directory: str, device: torch.device):
        last_ckpt = get_latest_checkpoint_path(directory)
        if last_ckpt:
            checkpoint.load(last_ckpt, device)
        else:
            logger.error(f"No checkpoints found in {directory}.")



from typing import Dict, List, Optional, Tuple

import numpy as np
import pandas as pd

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import DASK_MODULE_NAME
from ludwig.data.dataframe.base import DataFrameEngine
from ludwig.utils.types import DataFrame


@DeveloperAPI
def is_dask_lib(df_lib) -> bool:
    """Returns whether the dataframe library is dask."""
    return df_lib.__name__ == DASK_MODULE_NAME


@DeveloperAPI
def is_dask_backend(backend: Optional["Backend"]) -> bool:  # noqa: F821
    """Returns whether the backend's dataframe is dask."""
    return backend is not None and is_dask_lib(backend.df_engine.df_lib)


@DeveloperAPI
def is_dask_series_or_df(df: DataFrame, backend: Optional["Backend"]) -> bool:  # noqa: F821
    if is_dask_backend(backend):
        import dask.dataframe as dd

        return isinstance(df, dd.Series) or isinstance(df, dd.DataFrame)
    return False


@DeveloperAPI
def flatten_df(df: DataFrame, df_engine: DataFrameEngine) -> Tuple[DataFrame, Dict[str, Tuple]]:  # noqa: F821
    """Returns a flattened dataframe with a dictionary of the original shapes, keyed by dataframe columns."""
    # Workaround for: https://issues.apache.org/jira/browse/ARROW-5645
    column_shapes = {}
    for c in df.columns:
        df = df_engine.persist(df)
        shape = df_engine.compute(
            df_engine.map_objects(
                df[c],
                lambda x: np.array(x).shape,
            ).max()
        )

        if len(shape) > 1:
            column_shapes[c] = shape
            df[c] = df_engine.map_objects(df[c], lambda x: np.array(x).reshape(-1))
    return df, column_shapes


@DeveloperAPI
def unflatten_df(df: DataFrame, column_shapes: Dict[str, Tuple], df_engine: DataFrameEngine) -> DataFrame:  # noqa: F821
    """Returns an unflattened dataframe, the reverse of flatten_df."""
    for c in df.columns:
        shape = column_shapes.get(c)
        if shape:
            df[c] = df_engine.map_objects(df[c], lambda x: np.array(x).reshape(shape))
    return df


@DeveloperAPI
def to_numpy_dataset(df: DataFrame, backend: Optional["Backend"] = None) -> Dict[str, np.ndarray]:  # noqa: F821
    """Returns a dictionary of numpy arrays, keyed by the columns of the given dataframe."""
    dataset = {}
    for col in df.columns:
        res = df[col]
        if backend and is_dask_backend(backend):
            res = res.compute()
        if len(df.index) != 0:
            dataset[col] = np.stack(res.to_numpy())
        else:
            # Dataframe is empty.
            # Use to_list() directly, as np.stack() requires at least one array to stack.
            dataset[col] = res.to_list()
    return dataset


@DeveloperAPI
def from_numpy_dataset(dataset) -> pd.DataFrame:
    """Returns a pandas dataframe from the dataset."""
    col_mapping = {}
    for k, v in dataset.items():
        if len(v.shape) > 1:
            # unstacking, needed for ndarrays of dimension 2 and more
            (*vals,) = v
        else:
            # not unstacking. Needed because otherwise pandas casts types
            # the way it wants, like converting a list of float32 scalats
            # to a column of float64
            vals = v
        col_mapping[k] = vals
    return pd.DataFrame.from_dict(col_mapping)


@DeveloperAPI
def set_index_name(pd_df: pd.DataFrame, name: str) -> pd.DataFrame:
    pd_df.index.name = name
    return pd_df


@DeveloperAPI
def to_batches(df: pd.DataFrame, batch_size: int) -> List[pd.DataFrame]:
    return [df[i : i + batch_size].copy() for i in range(0, df.shape[0], batch_size)]


@DeveloperAPI
def from_batches(batches: List[pd.DataFrame]) -> pd.DataFrame:
    return pd.concat(batches)


@DeveloperAPI
def to_scalar_df(df: pd.DataFrame) -> pd.DataFrame:
    """Converts all columns in a pd.DataFrame to be scalar types.

    For object columns of lists, each element of the list is expanded into its own column named {column}_{index}. We
    assume all object columns are lists of the same length (i.e., tensor format output from preprocessing). It's also
    important that the relative order of the columns is preserved, to maintain consistency with other conversions like
    the one for Hummingbird.
    """
    scalar_df = df
    column_ordering = []
    for c, s in df.items():
        if s.dtype == "object":
            s_list = s.to_list()
            try:
                ncols = s_list[0].shape[0]
                split_cols = [f"{c}_{k}" for k in range(ncols)]
                sdf = pd.DataFrame(s_list, columns=split_cols)
                scalar_df = pd.concat([scalar_df, sdf], axis=1)
                column_ordering += split_cols
            except AttributeError as e:
                raise ValueError(f"Expected series of lists, but found {s_list[0]}") from e
        else:
            column_ordering.append(c)
    return scalar_df[column_ordering]



import logging
import os
import warnings
from typing import Any, Dict, List, Mapping, Optional, Tuple, TYPE_CHECKING, Union

import deepspeed
import deepspeed.comm
import torch
from deepspeed.utils.zero_to_fp32 import get_fp32_state_dict_from_zero_checkpoint
from packaging import version
from torch import nn
from torch.optim.optimizer import Optimizer

from ludwig.constants import MIN_POSSIBLE_BATCH_SIZE
from ludwig.distributed.ddp import DDPStrategy
from ludwig.modules.optimization_modules import get_optimizer_class_and_kwargs
from ludwig.utils.checkpoint_utils import Checkpoint
from ludwig.utils.model_utils import extract_tensors, replace_tensors

_deepspeed_0101 = version.parse(deepspeed.__version__) >= version.parse("0.10.1")


if TYPE_CHECKING:
    from ludwig.modules.lr_scheduler import LRScheduler
    from ludwig.schema.trainer import ECDTrainerConfig


DEFAULT_ZERO_OPTIMIZATION = {
    "stage": "auto",
    "stage3_gather_16bit_weights_on_model_save": "auto",
    "offload_optimizer": {"device": "auto"},
    "offload_param": {"device": "auto"},
}

# Filter out warnings about DeepSpeed use of deprecated methods. Can remove on upgrade to DeepSpeed 0.9.
warnings.filterwarnings(
    action="ignore",
    category=UserWarning,
    module="torch.distributed.distributed_c10d",
)


class DeepSpeedStrategy(DDPStrategy):
    def __init__(
        self,
        zero_optimization: Optional[Dict[str, Any]] = None,
        fp16: Optional[Dict[str, Any]] = None,
        bf16: Optional[Dict[str, Any]] = None,
        compression_training: Optional[Dict[str, Any]] = None,
        **kwargs
    ):
        # If we're initializing from a `deepspeed` CLI command, deepspeed will have already been initialized, as
        # indicated by the presence of the LOCAL_RANK var. Otherwise, we're initializing from Ray / torchrun, and will
        # need to set this var ourselves, then init DeepSpeed here.
        local_rank, local_size = os.environ.get("LOCAL_RANK"), os.environ.get("LOCAL_SIZE")
        init_deepspeed = local_rank is None or local_size is None

        super().__init__(**kwargs)
        self.zero_optimization = zero_optimization or DEFAULT_ZERO_OPTIMIZATION
        self.fp16 = fp16
        self.bf16 = bf16
        self.compression_training = compression_training

        if init_deepspeed:
            os.environ["LOCAL_RANK"] = str(self.local_rank())
            os.environ["LOCAL_SIZE"] = str(self.local_size())
            os.environ["RANK"] = str(self.rank())
            os.environ["WORLD_SIZE"] = str(self.size())
            deepspeed.init_distributed()

    def _log_on_init(self):
        logging.info("Using DeepSpeed strategy")

    def prepare(
        self,
        model: nn.Module,
        trainer_config: "ECDTrainerConfig",
        base_learning_rate: float,
    ) -> Tuple[nn.Module, Optimizer]:
        # If `batch_size=auto`, we set to MIN_POSSIBLE_BATCH_SIZE temporarily until auto-tuning adjusts it`
        # We can really set it to be whatever we want, as it will be overridden by the auto-tuning.
        batch_size = (
            trainer_config.batch_size if isinstance(trainer_config.batch_size, int) else MIN_POSSIBLE_BATCH_SIZE
        )
        # Paged and 8-bit optimizers are not supported by Deepspeed - just whatever is supported
        # by torch.optim.Optimizer. https://www.deepspeed.ai/docs/config-json/#optimizer-parameters.
        if trainer_config.optimizer.is_paged or trainer_config.optimizer.is_8bit:
            raise ValueError("Cannot use a paged or 8-bit optimizer with DeepSpeed.")
        optimizer_cls, optimizer_kwargs = get_optimizer_class_and_kwargs(trainer_config.optimizer, base_learning_rate)
        ds_config = {
            "amp": {
                "enabled": trainer_config.use_mixed_precision,
            },
            "optimizer": {"type": optimizer_cls.__name__, "params": optimizer_kwargs},
            "zero_optimization": self.zero_optimization,
            "gradient_clipping": trainer_config.gradient_clipping.clipglobalnorm,
            "train_micro_batch_size_per_gpu": batch_size,
            "gradient_accumulation_steps": trainer_config.gradient_accumulation_steps,
            "steps_per_print": trainer_config.steps_per_checkpoint or 10000,
        }

        # DeepSpeed doesn't like passing these params as None values
        if self.fp16 is not None:
            ds_config["fp16"] = self.fp16
        if self.bf16 is not None:
            ds_config["bf16"] = self.bf16
        if self.compression_training is not None:
            ds_config["compression_training"] = self.compression_training

        model_engine, optimizer, _, _ = deepspeed.initialize(
            model=model,
            model_parameters=model.parameters(),
            lr_scheduler=None,  # Don't let DeepSpeed manage the learning rate scheduler
            config=ds_config,
            dist_init_required=False,
        )

        if hasattr(optimizer, "optimizer"):
            # Zero-3 wraps the optimizer
            optimizer = optimizer.optimizer

        return model_engine, optimizer

    def prepare_for_inference(self, model: nn.Module) -> nn.Module:
        ds_config = {}
        model_engine = deepspeed.init_inference(model=model, config=ds_config)
        return model_engine

    def to_device(self, model: nn.Module, device: Optional[torch.device] = None) -> nn.Module:
        return model

    def backward(self, loss: torch.Tensor, model: nn.Module):
        # See: https://github.com/huggingface/accelerate/blob/main/src/accelerate/utils/deepspeed.py
        # runs backpropagation and handles mixed precision
        model.backward(loss)

        # Deepspeed's `engine.step` performs the following operations:
        # - gradient accumulation check
        # - gradient clipping
        # - optimizer step
        # - zero grad
        # - checking overflow
        # - lr_scheduler step (only if engine.lr_scheduler is not None)
        model.step()
        # and this plugin overrides the above calls with no-ops when Accelerate runs under
        # Deepspeed, but allows normal functionality for non-Deepspeed cases thus enabling a simple
        # training loop that works transparently under many training regimes.

    def step(self, optimizer: Optimizer, *args, **kwargs):
        # Handled by `self.backward(loss)`
        pass

    def zero_grad(self, optimizer: Optimizer):
        # Handled by `self.backward(loss)`
        pass

    def set_batch_size(self, model: nn.Module, batch_size: int):
        # Adapted from:
        # https://github.com/microsoft/DeepSpeed/blob/7ce371b139521b1ebbf052f0496b1a16397c1d19/deepspeed/runtime/engine.py#L422  # noqa: E501
        model._config.micro_batch_size_per_gpu = batch_size
        model._config.train_batch_size = batch_size * self.size() * model._config.gradient_accumulation_steps

    def barrier(self):
        deepspeed.comm.barrier()

    def allow_gradient_accumulation(self) -> bool:
        """DeepSpeed handles gradient accumulation internally."""
        return False

    def allow_mixed_precision(self) -> bool:
        """DeepSpeed handles mixed precision internally."""
        return False

    def allow_clip_gradients(self) -> bool:
        """DeepSpeed handles gradient clipping internally."""
        return False

    def prepare_before_load(self) -> bool:
        """DeepSpeed requires the engine to be re-initialized before loading.

        https://deepspeed.readthedocs.io/en/latest/model-checkpointing.html#loading-training-checkpoints
        """
        return True

    @classmethod
    def is_model_parallel(cls) -> bool:
        return True

    def create_checkpoint_handle(
        self,
        dist_model: nn.Module,
        model: nn.Module,
        optimizer: Optional[Optimizer] = None,
        scheduler: Optional["LRScheduler"] = None,
    ) -> Checkpoint:
        return DeepSpeedCheckpoint(self, dist_model, optimizer, scheduler)

    @classmethod
    def extract_model_for_serialization(cls, model: nn.Module) -> Union[nn.Module, Tuple[nn.Module, List[Dict]]]:
        return extract_tensors(model)

    @classmethod
    def replace_model_from_serialization(cls, state: Union[nn.Module, Tuple[nn.Module, List[Dict]]]) -> nn.Module:
        assert isinstance(state, tuple)
        model, model_weights = state
        replace_tensors(model, model_weights, torch.device("cpu"))
        return model


class DeepSpeedCheckpoint(Checkpoint):
    def prepare(self, directory: str):
        if self.distributed.local_rank() == 0:
            # Checkpoints need to be written on every rank, but the directory only needs to be created once per node.
            super().prepare(directory)

    def load(self, save_path: str, device: Optional[torch.device] = None) -> bool:
        """Load a checkpoint.

        For DeepSpeed, we need every worker to independently load back the model weights, as the checkpoints themselves
        may be sharded (when using DeepSpeed Zero3).

        https://deepspeed.readthedocs.io/en/latest/model-checkpointing.html#loading-training-checkpoints
        """
        _, client_state = self.model.load_checkpoint(save_path, load_lr_scheduler_states=False)
        self.global_step = self._get_global_step(client_state, save_path)
        if self.scheduler is not None and "scheduler_state" in client_state:
            self.scheduler.load_state_dict(client_state["scheduler_state"])
        return True

    def save(self, save_path: str, global_step: int):
        client_state = {
            "global_step": global_step,
        }
        if self.scheduler is not None:
            client_state["scheduler_state"] = self.scheduler.state_dict()

        kwargs = {}
        if _deepspeed_0101:
            kwargs["exclude_frozen_parameters"] = True

        self.model.save_checkpoint(save_path, client_state=client_state, **kwargs)

    def get_state_for_inference(self, save_path: str, device: Optional[torch.device] = None) -> Mapping[str, Any]:
        if self.model.zero_optimization_stage() == 3:
            return get_fp32_state_dict_from_zero_checkpoint(save_path)

        self.model.load_checkpoint(
            save_path, load_optimizer_states=False, load_lr_scheduler_states=False, load_module_only=True
        )
        return self.model.module.cpu().state_dict()



import json
import logging
import os
import string
from typing import Any, Dict, List, Optional, Set, Tuple, Type, TYPE_CHECKING

import pandas as pd

if TYPE_CHECKING:
    from ludwig.backend.base import Backend

from ludwig.models.retrieval import df_checksum, get_retrieval_model, RetrievalModel
from ludwig.utils.fs_utils import get_default_cache_location, makedirs, path_exists
from ludwig.utils.types import DataFrame, Series

logger = logging.getLogger(__name__)

CONTEXT = "__context__"
SAMPLE = "__sample__"
TASK = "__task__"

DEFAULT_ZERO_SHOT_PROMPT_TEMPLATE = """SAMPLE INPUT: {__sample__}

USER: Complete the following task: {__task__}

ASSISTANT:
"""


DEFAULT_FEW_SHOT_PROMPT_TEMPLATE = """Below is relevant context:

CONTEXT: {__context__}

CONTEXT is comprised of labeled samples whose embeddings were similar to that of the sample input. The labels in
these samples could aid you in your final prediction. Given this and no prior knowledge, follow the instructions
below.

SAMPLE INPUT: {__sample__}

USER: Complete the following task: {__task__}

ASSISTANT:
"""


def index_column(
    retrieval_config: Dict[str, Any],
    col_name: str,
    dataset_cols: Dict[str, Series],
    backend: "Backend",
    split_col: Optional[Series] = None,
) -> Tuple[RetrievalModel, str]:
    """Indexes a column for sample retrieval via embedding index lookup.

    This function indexes a column and saves the index artifact to disk. If an index name is provided as part of the
    `retrieval_config`, then the index in the ludwig cache with the corresponding name will be loaded instead of being
    built from scratch.

    To prevent data leakage, a split column must be provided. This ensures that the retrieval model only ever fetches
    samples from the training set.

    To ensure that the index is usable even if the original DataFrame is not available, the columns used to build the
    index are stored as part of the index.

    All operations in this function are performed on pandas objects, which means that you may run out of memory if your
    dataset is large.

    Args:
        retrieval_config (Dict[str, Any]): The retrieval config from the config object.
        col_name (str): The name of the column to index.
        dataset_cols (Dict[str, Series]): A dictionary mapping column names to their corresponding Series. `col_name`
            must be a key in this dictionary. These columns are stored as part of the index to ensure that the index
            is usable even if the original DataFrame is not available.
        df_engine (DataFrameEngine): The engine used to compute the columns into pandas objects.
        split_col (Optional[Series]): A column that indicates whether a sample is part of the training set. A sample
            is in the training set if the value in this column is 0.
    Returns:
        Tuple[RetrievalModel, str]: A tuple containing the retrieval model and the name of the index.
    """
    retrieval_model = get_retrieval_model(
        retrieval_config["type"],
        model_name=retrieval_config["model_name"],
    )

    index_name = retrieval_config["index_name"]
    index_cache_directory = os.path.join(get_default_cache_location(), "index")
    if not path_exists(index_cache_directory):
        makedirs(index_cache_directory, exist_ok=True)

    if index_name is None:
        if split_col is None:
            raise ValueError("split column must be provided if using retrieval")
        split_col = backend.df_engine.compute(split_col).astype(int)

        # TODO(geoffrey): add support for Dask DataFrames
        df = pd.DataFrame({name: backend.df_engine.compute(col) for name, col in dataset_cols.items()})
        df = df[split_col == 0]  # Ensures that the index is only built on the training set

        # Even if index name is not provided, we still want to check if an index for this df already exists in cache
        # If it does, load it and return immediately
        index_hash = df_checksum(df)
        index_name = f"embedding_index_{index_hash}"
        if path_exists(os.path.join(index_cache_directory, index_name)):
            logger.info(
                f"Index for this DataFrame with name '{index_name}' already exists. "
                f"Loading index from '{index_cache_directory}'"
            )
            retrieval_model.load_index(index_name, cache_directory=index_cache_directory)
            return retrieval_model, index_name

        # Build index if index name is not provided and index for this df does not already exist in cache
        retrieval_model.create_dataset_index(df, backend, columns_to_index=[col_name])
        logger.info(f"Saving index to cache directory '{index_cache_directory}' with name '{index_name}'")
        retrieval_model.save_index(index_name, cache_directory=index_cache_directory)
    else:
        logger.info(f"Loading index from cache directory '{index_cache_directory}' with name '{index_name}'")
        retrieval_model.load_index(index_name, cache_directory=index_cache_directory)
    return retrieval_model, index_name


def format_input_with_prompt(
    input_col_name: str,
    dataset_df: DataFrame,
    backend: "Backend",
    task_str: str,
    retrieval_model: Optional[RetrievalModel] = None,
    k: int = -1,
    template: Optional[str] = None,
) -> Series:
    """Returns a new Series with the input column data formatted with the prompt.

    A prompt can either be zero-shot or few-shot. A zero-shot prompt is comprised of some (unlabeled) input and a task
    to be completed given the input. A few-shot prompt additionally includes some dynamically retrieved context, which
    is retrieved using the `retrieval_model.search` function.

    A template can be provided to customize the prompt. The template must be a string with the following fields:
        - __sample__ or at least one column from the input dataset: The input sample.
        - __context__: The context retrieved by the `search_fn` function. Only required if `search_fn` is provided.
        - __task__: The task to be completed given the input. Only required if `task` is set in the prompt config.

    Zero-shot example:

    Before formatting:

        input_col = ["I am happy"]
        task_str = "sentiment analysis"

    After formatting:

        input_col = ["SAMPLE INPUT: I am happy\n\nUSER: Complete the following task: sentiment analysis\n\nASSISTANT:"]

    Args:
        input_col_name (str): The name of the input column.
        dataset_df (DataFrame): The input dataset.
        backend (Backend): The backend used for map operations.
        task_str (str): The task to be completed given the input.
        retrieval_model (Optional[RetrievalModel]): The retrieval model used to retrieve context. If provided, the
            prompt will be few-shot. If not provided, the prompt will be zero-shot.
        k (int): The number of samples to retrieve. Only required if `retrieval_model` is provided.
        template (Optional[str]): The template to use for the prompt. If not provided, the default will be used.

    Returns:
        Series: A new Series with the input column data formatted with the prompt.
    """
    # determine if this is a few-shot or zero-shot prompt
    # few-shot prompts require a search function that returns samples from some dataset
    is_few_shot = retrieval_model is not None

    # if no template is provided, use the default template
    if template is None:
        if is_few_shot:
            template = DEFAULT_FEW_SHOT_PROMPT_TEMPLATE
        else:
            template = DEFAULT_ZERO_SHOT_PROMPT_TEMPLATE

    # ensure that the prompt template has all required fields
    template_fields, field_to_dtype = _get_template_fields(template)
    try:
        _validate_prompt_template(template_fields, task_str, is_few_shot, dataset_df.columns, input_col_name)
    except ValueError as e:
        raise ValueError(f"template invalid for {'few-shot' if is_few_shot else 'zero-shot'} prompt: {e}")

    def generate_prompt(df: pd.DataFrame):
        if CONTEXT in template_fields:
            df[CONTEXT] = retrieval_model.search(df, backend, k=k, return_data=True)
        if SAMPLE in template_fields:
            # During preprocessing, we're inserting quotes that change the token IDs completely if we
            # don't remove the " from the string. For parity with expected user output, we need to get rid of them.
            # TODO(Arnav): see if there's a way to only remove them if the entry does't have quotes. This currently
            # removes all " from the string (even those not added by json.dumps), which is not ideal.
            df[SAMPLE] = df[input_col_name].map(lambda entry: json.dumps(entry, indent=2).strip('"'))
        if TASK in template_fields:
            df[TASK] = task_str

        def generate_prompt_for_row(row):
            kwargs = {col: field_to_dtype[col](row[col]) for col in template_fields}
            return template.format(**kwargs)

        return df.apply(generate_prompt_for_row, axis=1)

    result = backend.df_engine.map_partitions(dataset_df, generate_prompt, meta=(input_col_name, "object"))
    result = backend.df_engine.persist(result)  # persist to prevent re-computation
    return result


def _validate_prompt_template(
    template_fields: Set[str], task: Optional[str], is_few_shot: bool, columns: List[str], input_col_name: str
):
    """Validates that the template contains the necessary fields for the prompt."""
    if is_few_shot and CONTEXT not in template_fields:
        raise ValueError(f"Prompt template must contain the '{CONTEXT}' field for few-shot learning")

    if task is not None and TASK not in template_fields:
        raise ValueError(f"Prompt template must contain the '{TASK}' field if a task is provided")

    if SAMPLE in template_fields:
        if input_col_name not in columns:
            raise ValueError(
                f"Prompt template contains the '{SAMPLE}' field, "
                f"but the input column '{input_col_name}' is not in the dataset"
            )
    elif not any(col in template_fields for col in columns):
        raise ValueError(
            f"Prompt template must contain either the '{SAMPLE}' field or one of the columns from the dataset"
        )


def _get_template_fields(template: str) -> Tuple[Set[str], Dict[str, Type]]:
    """Returns the fields in the template."""
    parsed = [t for t in string.Formatter().parse(template) if t[1] is not None]
    field_set = {field for _, field, _, _ in parsed}
    dtype_map = {field: _get_dtype(format_spec) for _, field, format_spec, _ in parsed}
    return field_set, dtype_map


def _get_dtype(format_spec: str) -> Type:
    # We need to prepare data in the row for different formatting options.
    # If you have a number like 0.1234 in the DF and you want to format it like {number:.2f} it will fail if the
    # number is represented as a string in the DF. So we need to cast it to a float before formatting.
    if not format_spec:
        return str
    if "f" in format_spec:
        return float
    raise ValueError(f"Unsupported template format spec: {format_spec}")



import datetime
from contextlib import nullcontext as does_not_raise
from typing import Any, ContextManager

import pytest

from ludwig.utils.date_utils import convert_number_to_datetime


@pytest.fixture(scope="module")
def reference_datetime() -> datetime.datetime:
    return datetime.datetime.utcfromtimestamp(1691600953.443032)


@pytest.mark.parametrize(
    "timestamp,raises",
    [
        pytest.param(1691600953.443032, does_not_raise(), id="float-s"),
        pytest.param(1691600953443.032, does_not_raise(), id="float-ms"),
        pytest.param(1691600953, does_not_raise(), id="int-s"),
        pytest.param(1691600953443, does_not_raise(), id="int-ms"),
        pytest.param("1691600953.443032", does_not_raise(), id="string[float]-s"),
        pytest.param("1691600953443.0032", does_not_raise(), id="string[float]-ms"),
        pytest.param("1691600953", does_not_raise(), id="string[int]-s"),
        pytest.param("1691600953443", does_not_raise(), id="string[int]-ms"),
        pytest.param("foo", pytest.raises(ValueError), id="string error"),
        pytest.param([1691600953.443032], pytest.raises(ValueError), id="list error"),
        pytest.param(datetime.datetime(2023, 8, 9, 13, 9, 13), pytest.raises(ValueError), id="datetime error"),
        pytest.param(None, pytest.raises(ValueError), id="NoneType error"),
    ],
)
def test_convert_number_to_datetime(reference_datetime: datetime.datetime, timestamp: Any, raises: ContextManager):
    """Ensure that numeric timestamps are correctly converted to datetime objects.

    Args:
        reference_datetime: A datetime object with the expected date/time
        timestamp: The timestamp to convert in s or ms
        raises: context manager to check for expected exceptions
    """
    with raises:
        dt = convert_number_to_datetime(timestamp)

        # Check that the returned datetime is accurate to the scale of seconds.
        assert dt.year == reference_datetime.year
        assert dt.month == reference_datetime.month
        assert dt.day == reference_datetime.day
        assert dt.hour == reference_datetime.hour
        assert dt.minute == reference_datetime.minute
        assert dt.second == reference_datetime.second



import argparse
import logging
import sys
from typing import Optional

from ludwig.utils.print_utils import get_logging_level_registry
from ludwig.utils.upload_utils import HuggingFaceHub

logger = logging.getLogger(__name__)


def get_upload_registry():
    return {
        "hf_hub": HuggingFaceHub,
    }


def upload_cli(
    service: str,
    repo_id: str,
    model_path: str,
    repo_type: str = "model",
    private: bool = False,
    commit_message: str = "Upload trained [Ludwig](https://ludwig.ai/latest/) model weights",
    commit_description: Optional[str] = None,
    **kwargs,
) -> None:
    """Create an empty repo on the HuggingFace Hub and upload trained model artifacts to that repo.

    Args:
        service (`str`):
            Name of the hosted model service to push the trained artifacts to.
            Currently, this only supports `hf_hub`.
        repo_id (`str`):
            A namespace (user or an organization) and a repo name separated
            by a `/`.
        model_path (`str`):
            The path of the saved model. This is the top level directory where
            the models weights as well as other associated training artifacts
            are saved.
        private (`bool`, *optional*, defaults to `False`):
            Whether the model repo should be private.
        repo_type (`str`, *optional*):
            Set to `"dataset"` or `"space"` if uploading to a dataset or
            space, `None` or `"model"` if uploading to a model. Default is
            `None`.
        commit_message (`str`, *optional*):
            The summary / title / first line of the generated commit. Defaults to:
            `f"Upload {path_in_repo} with huggingface_hub"`
        commit_description (`str` *optional*):
            The description of the generated commit
    """
    model_service = get_upload_registry().get(service, "hf_hub")
    hub = model_service()
    hub.login()
    hub.upload(
        repo_id=repo_id,
        model_path=model_path,
        repo_type=repo_type,
        private=private,
        commit_message=commit_message,
        commit_description=commit_description,
    )


def cli(sys_argv):
    parser = argparse.ArgumentParser(
        description="This script pushes a trained model to a hosted model repository service",
        prog="ludwig upload",
        usage="%(prog)s [options]",
    )

    # ---------------
    # Required parameters
    # ---------------
    parser.add_argument(
        "service",
        help="Name of the model repository service.",
        default="hf_hub",
        choices=["hf_hub"],
    )

    parser.add_argument(
        "-r",
        "--repo_id",
        help="Name of the repo. This will be created if it doesn't exist. Format: username/repo_name",
        required=True,
    )

    parser.add_argument("-m", "--model_path", help="Path of the trained model on disk", required=True)

    # ---------------
    # Optional parameters
    # ---------------
    parser.add_argument("-p", "--private", help="Make the repo private", default=False, choices=[True, False])

    parser.add_argument(
        "-t", "--repo_type", help="Type of repo", default="model", choices=["model", "space", "dataset"]
    )

    parser.add_argument(
        "-c",
        "--commit_message",
        help="The summary / title / first line of the generated commit.",
        default="Upload trained [Ludwig](https://ludwig.ai/latest/) model weights",
    )

    parser.add_argument("-d", "--commit_description", help="The description of the generated commit", default=None)

    parser.add_argument(
        "-l",
        "--logging_level",
        default="info",
        help="The level of logging to use",
        choices=["critical", "error", "warning", "info", "debug", "notset"],
    )

    args = parser.parse_args(sys_argv)

    args.logging_level = get_logging_level_registry()[args.logging_level]
    logging.getLogger("ludwig").setLevel(args.logging_level)
    global logger
    logger = logging.getLogger("ludwig.upload")

    upload_cli(**vars(args))


if __name__ == "__main__":
    cli(sys.argv[1:])



import os

import numpy as np
import pytest

from ludwig.api import LudwigModel
from ludwig.constants import MODEL_ECD, MODEL_GBM, PREPROCESSING, PROC_COLUMN, TRAINER
from tests.integration_tests.test_gbm import category_feature
from tests.integration_tests.utils import binary_feature, generate_data, number_feature, run_test_suite, text_feature


@pytest.mark.slow
@pytest.mark.parametrize(
    "backend",
    [
        pytest.param("local", id="local"),
        pytest.param("ray", id="ray", marks=pytest.mark.distributed),
    ],
)
def test_onehot_encoding(tmpdir, backend, ray_cluster_2cpu):
    input_features = [
        number_feature(),
        category_feature(encoder={"type": "onehot"}),
    ]
    output_features = [binary_feature()]

    data_csv_path = os.path.join(tmpdir, "dataset.csv")
    dataset = generate_data(input_features, output_features, data_csv_path)
    config = {"input_features": input_features, "output_features": output_features, TRAINER: {"epochs": 2}}
    run_test_suite(config, dataset, backend)


@pytest.mark.slow
@pytest.mark.parametrize(
    "backend",
    [
        pytest.param("local", id="local"),
        pytest.param("ray", id="ray", marks=pytest.mark.distributed),
    ],
)
def test_hf_text_embedding(tmpdir, backend, ray_cluster_2cpu):
    input_features = [
        number_feature(),
        text_feature(
            encoder={
                "type": "auto_transformer",
                "pretrained_model_name_or_path": "hf-internal-testing/tiny-bert-for-token-classification",
            },
            preprocessing={"cache_encoder_embeddings": True},
        ),
    ]
    output_features = [binary_feature()]

    data_csv_path = os.path.join(tmpdir, "dataset.csv")
    dataset = generate_data(input_features, output_features, data_csv_path)

    config = {"input_features": input_features, "output_features": output_features, TRAINER: {"epochs": 1}}
    run_test_suite(config, dataset, backend)


@pytest.mark.slow
@pytest.mark.parametrize("cache_encoder_embeddings", [True, False, None])
@pytest.mark.parametrize("model_type", [MODEL_ECD, MODEL_GBM])
def test_onehot_encoding_preprocessing(model_type, cache_encoder_embeddings, tmpdir):
    vocab_size = 5
    input_features = [
        category_feature(encoder={"type": "onehot", "vocab_size": vocab_size}),
        number_feature(),
    ]
    output_features = [binary_feature()]

    if cache_encoder_embeddings is not None:
        if PREPROCESSING not in input_features[0]:
            input_features[0][PREPROCESSING] = {}
        input_features[0][PREPROCESSING]["cache_encoder_embeddings"] = cache_encoder_embeddings

    # Need sufficiently high number of examples to ensure at least one of each category type appears
    data_csv_path = os.path.join(tmpdir, "dataset.csv")
    num_examples = 100
    dataset_fp = generate_data(input_features, output_features, data_csv_path, num_examples)
    config = {
        "model_type": model_type,
        "input_features": input_features,
        "output_features": output_features,
    }

    # Run preprocessing
    ludwig_model = LudwigModel(config, backend="local")
    proc_dataset = ludwig_model.preprocess(training_set=dataset_fp)

    # Check preprocessed output
    proc_df = ludwig_model.backend.df_engine.compute(proc_dataset.training_set.to_df())
    proc_col = input_features[0][PROC_COLUMN]
    proc_series = proc_df[proc_col]

    # GBMs always cache embeddings, ECD will not by default, but will if set to `cache_encoder_embeddings=true`
    expected_cache_encoder_embeddings = (cache_encoder_embeddings or False) if model_type == MODEL_ECD else True
    if expected_cache_encoder_embeddings:
        assert proc_series.values.dtype == "object"
        data = np.stack(proc_series.values)
        assert data.shape == (num_examples, vocab_size)

        # Only one element in each row should be 1
        assert all(x == 1 for x in data.sum(axis=1))
    else:
        assert proc_series.values.dtype == "int8"
        data = proc_series.to_numpy()
        assert data.shape == (num_examples,)


def test_hf_text_embedding_tied(tmpdir):
    input_features = [
        text_feature(
            encoder={
                "type": "auto_transformer",
                "pretrained_model_name_or_path": "hf-internal-testing/tiny-bert-for-token-classification",
            },
            preprocessing={"cache_encoder_embeddings": True},
        ),
        text_feature(
            encoder={
                "type": "auto_transformer",
                "pretrained_model_name_or_path": "hf-internal-testing/tiny-bert-for-token-classification",
            },
            preprocessing={"cache_encoder_embeddings": True},
        ),
    ]
    input_features[1]["tied"] = input_features[0]["name"]
    output_features = [binary_feature()]

    data_csv_path = os.path.join(tmpdir, "dataset.csv")
    dataset = generate_data(input_features, output_features, data_csv_path)

    config = {"input_features": input_features, "output_features": output_features, TRAINER: {"epochs": 1}}
    run_test_suite(config, dataset, "local")



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import time
from datetime import date, datetime
from typing import Union

import numpy as np
from dateutil.parser import parse, ParserError

from ludwig.api_annotations import DeveloperAPI

SCALE_S = np.floor(np.log10(time.time()))


@DeveloperAPI
def create_vector_from_datetime_obj(datetime_obj):
    yearday = datetime_obj.toordinal() - date(datetime_obj.year, 1, 1).toordinal() + 1

    midnight = datetime_obj.replace(hour=0, minute=0, second=0, microsecond=0)
    second_of_day = (datetime_obj - midnight).seconds

    return [
        datetime_obj.year,
        datetime_obj.month,
        datetime_obj.day,
        datetime_obj.weekday(),
        yearday,
        datetime_obj.hour,
        datetime_obj.minute,
        datetime_obj.second,
        second_of_day,
    ]


@DeveloperAPI
def parse_datetime(timestamp: Union[float, int, str]) -> datetime:
    """Parse a datetime from a string or a numeric timestamp.

    Args:
        timestamp: A datetime string or numeric timestamp.

    Returns:
        A datetime representation of `timestamp`.
    """
    try:
        dt = parse(timestamp)
    except (OverflowError, ParserError, TypeError):
        dt = convert_number_to_datetime(timestamp)

    return dt


@DeveloperAPI
def convert_number_to_datetime(timestamp: Union[float, int, str]) -> datetime:
    """Convert a numeric timestamp to a datetime object.

    `datetime` objects can be created from POSIX timestamps like those returned by `time.time()`.

    Args:
        timestamp: A numeric timestamp.

    Returns:
        A datetime representation of `timestamp`.

    Raises:
        ValueError: Raised if `timestamp` is not a number or not a valid datetime.
    """
    try:
        timestamp = float(timestamp)
    except TypeError:
        raise ValueError(f"Provided value {timestamp} is not a valid numeric timestamp")

    # Determine the unit of the timestamp
    ts_scale = np.floor(np.log10(timestamp))

    # `datetime.datetime.fromtimestamp` expects a timestamp in seconds. Rescale the timestamp if it is not in seconds.
    if SCALE_S < ts_scale:
        delta = ts_scale - SCALE_S
        timestamp = timestamp / np.power(10, delta)

    # Convert the timestamp to a datetime object. If it is not a valid timestamp, `ValueError` is raised.
    dt = datetime.utcfromtimestamp(timestamp)
    return dt



from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import DATE, MODEL_ECD
from ludwig.schema import utils as schema_utils
from ludwig.schema.encoders.base import BaseEncoderConfig
from ludwig.schema.encoders.utils import EncoderDataclassField
from ludwig.schema.features.base import BaseInputFeatureConfig
from ludwig.schema.features.preprocessing.base import BasePreprocessingConfig
from ludwig.schema.features.preprocessing.utils import PreprocessingDataclassField
from ludwig.schema.features.utils import ecd_defaults_config_registry, ecd_input_config_registry, input_mixin_registry
from ludwig.schema.utils import BaseMarshmallowConfig, ludwig_dataclass


@DeveloperAPI
@ecd_defaults_config_registry.register(DATE)
@input_mixin_registry.register(DATE)
@ludwig_dataclass
class DateInputFeatureConfigMixin(BaseMarshmallowConfig):
    """DateInputFeatureConfigMixin is a dataclass that configures the parameters used in both the date input
    feature and the date global defaults section of the Ludwig Config."""

    preprocessing: BasePreprocessingConfig = PreprocessingDataclassField(feature_type=DATE)

    encoder: BaseEncoderConfig = EncoderDataclassField(
        MODEL_ECD,
        feature_type=DATE,
        default="embed",
    )


@DeveloperAPI
@ecd_input_config_registry.register(DATE)
@ludwig_dataclass
class DateInputFeatureConfig(DateInputFeatureConfigMixin, BaseInputFeatureConfig):
    """DateInputFeature is a dataclass that configures the parameters used for a date input feature."""

    type: str = schema_utils.ProtectedString(DATE)



from typing import List, Type, Union

from marshmallow.fields import Field

from ludwig.api_annotations import DeveloperAPI
from ludwig.schema import utils as schema_utils
from ludwig.schema.hyperopt.utils import register_parameter_config
from ludwig.schema.utils import ludwig_dataclass


def quantization_number_field(dtype: Union[Type[float], Type[int]] = float, default=None) -> Field:
    description = (
        "Quantization number. Output values will be rounded to the nearest increment of `q` in range."
        "Quantization makes the upper bound inclusive."
    )
    if dtype is int:
        field = schema_utils.Integer(default=default, allow_none=True, description=description)
    else:
        field = schema_utils.FloatRange(default=default, allow_none=True, description=description)

    return field


def log_base_field(default: float = 10) -> Field:
    return schema_utils.FloatRange(default=default, description="Logarithmic base.")


@DeveloperAPI
@register_parameter_config("choice")
@ludwig_dataclass
class ChoiceParameterConfig(schema_utils.BaseMarshmallowConfig):
    """Config for a randomly sampled categorical search space."""

    space: str = schema_utils.ProtectedString("choice")

    categories: List = schema_utils.OneOfOptionsField(
        default=None,
        allow_none=True,
        description=(
            "The list of values to use in creating the categorical space. The type of each value of the list is "
            "general, i.e., they could be strings, integers, floats and anything else, even entire dictionaries."
        ),
        field_options=[
            schema_utils.List(list_type=float, allow_none=False, description="The list of floats to randomly sample."),
            schema_utils.List(list_type=int, allow_none=False, description="The list of integers to randomly sample."),
            schema_utils.List(list_type=str, allow_none=False, description="The list of strings to randomly sample."),
            schema_utils.List(
                list_type=list,
                inner_type=dict,
                allow_none=False,
                description="The list of lists of configs to randomly sample.",
            ),
            schema_utils.DictList(allow_none=False, description="A list of nested config parameters to sample."),
        ],
    )


@DeveloperAPI
@register_parameter_config("grid_search")
@ludwig_dataclass
class GridSearchParameterConfig(schema_utils.BaseMarshmallowConfig):
    """Config for a grid search space."""

    space: str = schema_utils.ProtectedString("grid_search")

    values: List = schema_utils.OneOfOptionsField(
        default=None,
        allow_none=True,
        description=(
            "The list of values to use in creating the grid search space. The type of each value of the list is "
            "general, i.e., they could be strings, integers, floats and anything else, even entire dictionaries."
        ),
        field_options=[
            schema_utils.List(list_type=float, allow_none=False, description="The list of floats to randomly sample."),
            schema_utils.List(list_type=int, allow_none=False, description="The list of integers to randomly sample."),
            schema_utils.List(list_type=str, allow_none=False, description="The list of strings to randomly sample."),
        ],
    )


@DeveloperAPI
@register_parameter_config("uniform")
@ludwig_dataclass
class UniformParameterConfig(schema_utils.BaseMarshmallowConfig):
    """Config for a real-valued uniform search space."""

    space: str = schema_utils.ProtectedString("uniform")

    lower: float = schema_utils.FloatRange(default=None, description="The minimum value the parameter can have.")

    upper: float = schema_utils.FloatRange(default=None, description="The maximum value the parameter can have.")


@DeveloperAPI
@register_parameter_config("quniform")
@ludwig_dataclass
class QUniformParameterConfig(UniformParameterConfig):
    """Config for a real-valued uniform search space with quantization."""

    space: str = schema_utils.ProtectedString("quniform")

    q: float = quantization_number_field()


@DeveloperAPI
@register_parameter_config("loguniform")
@ludwig_dataclass
class LogUniformParameterConfig(UniformParameterConfig):
    """Config for a log-scaled real-valued uniform numeric search space."""

    space: str = schema_utils.ProtectedString("loguniform")

    base: float = log_base_field()


@DeveloperAPI
@register_parameter_config("qloguniform")
@ludwig_dataclass
class QLogUniformParameterConfig(UniformParameterConfig):
    """Config for a log-scaled real-valued uniform search space with quantization."""

    space: str = schema_utils.ProtectedString("qloguniform")

    q: float = quantization_number_field()

    base: float = log_base_field()


@DeveloperAPI
@register_parameter_config("randn")
@ludwig_dataclass
class RandnParameterConfig(schema_utils.BaseMarshmallowConfig):
    """Config for a Gaussian search space."""

    space: str = schema_utils.ProtectedString("randn")

    mean: float = schema_utils.FloatRange(default=0.0, description="Mean of the  normal distribution.")

    sd: float = schema_utils.FloatRange(default=1.0, description="Standard deviation of the normal distribution.")


@DeveloperAPI
@register_parameter_config("qrandn")
@ludwig_dataclass
class QRandnParameterConfig(RandnParameterConfig):
    """Config for a Gaussian search space with quantization."""

    space: str = schema_utils.ProtectedString("qrandn")

    q: float = quantization_number_field()


@DeveloperAPI
@register_parameter_config("randint")
@ludwig_dataclass
class RandintParameterConfig(schema_utils.BaseMarshmallowConfig):
    """Config for an integer-valued uniform search space."""

    space: str = schema_utils.ProtectedString("randint")

    lower: int = schema_utils.Integer(default=None, description="The minimum value the parameter can have.")

    upper: int = schema_utils.Integer(default=None, description="The maximum value the parameter can have.")


@DeveloperAPI
@register_parameter_config("qrandint")
@ludwig_dataclass
class QRandintParameterConfig(RandintParameterConfig):
    """Config for an integer-valued uniform search space with quantization."""

    space: str = schema_utils.ProtectedString("qrandint")

    q: int = quantization_number_field(dtype=int)


@DeveloperAPI
@register_parameter_config("lograndint")
@ludwig_dataclass
class LogRandintParameterConfig(RandintParameterConfig):
    """Config for an log-scaled integer-valued search space."""

    space: str = schema_utils.ProtectedString("lograndint")

    base: float = log_base_field()


@DeveloperAPI
@register_parameter_config("qlograndint")
@ludwig_dataclass
class QLogRandintParameterConfig(RandintParameterConfig):
    """Config for an log-scaled integer-valued search space with quantization."""

    space: str = schema_utils.ProtectedString("qlograndint")

    q: int = quantization_number_field(dtype=int)

    base: float = log_base_field()
